# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from future.utils import string_types

import datetime
import math

import requests

import blackfynn.log as log
from blackfynn.api.base import APIBase
from blackfynn.extensions import require_extension, pandas as pd, numpy as np
from blackfynn.models import (
    BaseDataNode,
    Collection,
    Dataset,
    File,
    Organization,
    Tabular,
    TabularSchema,
    TabularSchemaColumn,
    User,
    StatusLogResponse,
    StatusLogEntry,
    PublishInfo,
    UserCollaborator,
    TeamCollaborator,
)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Dataset
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class DatasetsAPI(APIBase):
    """
    Interface for managing datasets on the platform.
    """
    base_uri='/datasets'
    name = 'datasets'

    def get(self, ds):
        id = self._get_id(ds)
        resp = self._get( self._uri('/{id}', id=id))
        return Dataset.from_dict(resp, api=self.session)

    def published(self,ds):
        id = self._get_id(ds)
        resp = self._get( self._uri('/{id}/published', id=id))
        if resp['status'] == 'PUBLISH_SUCCEEDED':
            resp['latest_doi'] = self._get( self._uri('/{id}/doi', id=id))["doi"]
        return PublishInfo.from_dict(resp)

    def package_count(self,ds):
        id = self._get_id(ds)
        resp = self._get( self._uri('/{id}/packageTypeCounts', id=id))
        file_count = 0
        for key, value in resp.items():
            file_count += value
        return file_count

    def status_log(self,ds,limit,offset):
        id = self._get_id(ds)
        resp = self._get( self._uri('/{id}/status-log?limit={limit}&offset={offset}', id=id, limit=limit, offset=offset))
        return StatusLogResponse.from_dict(resp)

    def team_collaborators(self,ds):
        id = self._get_id(ds)
        resp = self._get( self._uri('/{id}/collaborators/teams', id=id))
        return [TeamCollaborator.from_dict(t) for t in resp]

    def user_collaborators(self,ds):
        id = self._get_id(ds)
        resp = self._get( self._uri('/{id}/collaborators/users', id=id))
        return [UserCollaborator.from_dict(u) for u in resp]

    def owner(self, ds):

        return next(iter(filter(lambda x: x.role == 'owner', self.user_collaborators(ds))))

    def get_by_name_or_id(self, name_or_id):
        """
        Get Dataset by name or ID.

        When using name, this ignores case, spaces, hyphens, and underscores
        such that these are equivalent:

          - "My Dataset"
          - "My-dataset"
          - "mydataset"
          - "my_DataSet"
          - "mYdata SET"

        """
        def name_key(n):
            return n.lower().strip().replace(' ', '').replace('_','').replace('-','')

        search_key = name_key(name_or_id)

        def is_match(ds):
            return (name_key(ds.name) == search_key) or (ds.id == name_or_id)

        matches = [ds for ds in self.get_all() if is_match(ds)]
        return matches[0] if matches else None

    def get_all(self):
        resp = self._get( self._uri('/'))
        return [Dataset.from_dict(ds, api=self.session) for ds in resp]

    def create(self, ds):
        """
        Create a dataset on the platform
        """
        if self.get_by_name_or_id(ds.name) is not None:
            raise Exception("Dataset with name {} already exists".format(ds.name))

        resp = self._post('', json=ds.as_dict())
        return Dataset.from_dict(resp, api=self.session)

    def update(self, ds):
        """
        Update a dataset on the platform
        """
        id = self._get_id(ds)
        resp = self._put( self._uri('/{id}', id=id), json=ds.as_dict())
        return Dataset.from_dict(resp, api=self.session)

    def delete(self, ds):
        """
        Delete a dataset on the platform
        """
        id = self._get_id(ds)
        resp = self._del( self._uri('/{id}', id=id))
        ds.id = None
        return resp


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Data
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class DataAPI(APIBase):
    """
    Interface for lower-level data operations on the platform.
    """
    base_uri='/data'
    name = 'data'

    def update_properties(self, thing):
        """
        Update properties for an object/package on the platform.
        """
        path = self._uri('/{id}/properties', id=thing.id)
        body = {"properties": [m.as_dict() for m in thing.properties]}

        return self._put(path, json=body)

    def delete(self, *things):
        """
        Deletes objects from the platform
        """
        ids = list(set([self._get_id(x) for x in things]))
        r = self._post('/delete', json=dict(things=ids))
        if len(r['success']) != len(ids):
            failures = [f['id'] for f in r['failures']]
            print("Unable to delete objects: {}".format(failures))

        for thing in things:
            if isinstance(thing, BaseDataNode):
                thing.id = None

        return r

    def move(self, destination, *things):
        """
        Moves objects to the destination package
        """
        ids = [self._get_id(x) for x in things]
        # if destination is None, things will get moved into their containing dataset
        dest = self._get_id(destination) if destination is not None else None
        return self._post("/move", json=dict(things=ids, destination=dest))

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Packages
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class PackagesAPI(APIBase):
    """
    Interface for task/workflow objects on Blackfynn platform
    """
    base_uri = '/packages'
    name = 'packages'

    def create(self, pkg):
        """
        Create data package on platform
        """
        resp = self._post('', json=pkg.as_dict())
        pkg = self._get_package_from_data(resp)
        return pkg

    def update(self, pkg, **kwargs):
        """
        Update package on platform
        """
        d = pkg.as_dict()
        d.update(kwargs)
        d.pop("state",None)
        resp = self._put(self._uri('/{id}', id=pkg.id), json=d)
        pkg = self._get_package_from_data(resp)
        return pkg

    def get(self, pkg, include=None):
        """
        Get package object

        pkg:     can be DataPackage ID or DataPackage object.
        include: list of fields to force-include in response (if available)
        """
        pkg_id = self._get_id(pkg)

        params = None
        if include is not None:
            if isinstance(include, string_types):
                params = {'include': include}
            if hasattr(include, '__iter__'):
                params = {'include': ','.join(include)}

        resp = self._get(self._uri('/{id}', id=pkg_id), params=params)

        # TODO: cast to specific DataPackages based on `type`
        pkg = self._get_package_from_data(resp)
        return pkg

    def process(self, pkg):
        """
        Process a package that has been successfully uploaded but not yet processed
        """
        try:
            self._put(self._uri('/{id}/process', id=pkg.id))
            return True
        except requests.exceptions.HTTPError as error:
            response = error.response
            status_code = response.status_code
            message = response.json().get('message')

            if status_code == requests.codes.bad_request and message is not None:
              raise Exception(message)
            else:
              raise error

    def get_sources(self, pkg):
        """
        Returns the sources of a DataPackage. Sources are the raw, unmodified
        files (if they exist) that contains the package's data.
        """
        pkg_id = self._get_id(pkg)
        resp = self._get(self._uri('/{id}/sources', id=pkg_id))
        for r in resp:
            r['content'].update(dict(pkg_id=pkg_id))

        return [File.from_dict(r, api=self.session) for r in resp]

    def get_files(self, pkg):
        """
        Returns the files of a DataPackage. Files are the possibly modified
        source files (e.g. converted to a different format), but they could also
        be the source files themselves.
        """
        pkg_id = self._get_id(pkg)
        resp = self._get(self._uri('/{id}/files', id=pkg_id))
        for r in resp:
            r['content'].update(dict(pkg_id=pkg_id))

        return [File.from_dict(r, api=self.session) for r in resp]

    def get_view(self, pkg):
        """
        Returns the object(s) used to view the package. This is typically a set of
        file objects, that may be the DataPackage's sources or files, but could also be
        a unique object specific for the viewer.
        """
        pkg_id = self._get_id(pkg)
        resp = self._get(self._uri('/{id}/view', id=pkg_id))
        for r in resp:
            r['content'].update(dict(pkg_id=pkg_id))

        return [File.from_dict(r, api=self.session) for r in resp]

    def get_presigned_url_for_file(self, pkg, file):
        args = dict(
            pkg_id = self._get_id(pkg),
            file_id = self._get_id(file)
        )
        resp = self._get(self._uri('/{pkg_id}/files/{file_id}', **args))
        if 'url' in resp:
            return resp['url']
        else:
            raise Exception("Unable to get URL for file ID = {}".format(file_id))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Tabular
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class TabularAPI(APIBase):
    base_uri = "/tabular"
    name = 'tabular'

    def _get_data_chunked(self, package, chunk_size, offset, order_by, order_direction):
        path = self._uri('/{id}', id=self._get_id(package))

        params = dict(limit = chunk_size, offset = offset, order_direction = order_direction)
        if order_by is not None:
            params['orderBy'] = order_by

        return self._get(path, params=params)

    @require_extension
    def get_tabular_data_iter(self, package, offset, order_by, order_direction, chunk_size=10000):
        """
        Return iterator that yields chunk_size data each call
        """
        if chunk_size > 10000:
            raise ValueError('Chunk size must be less than 10000')

        schema = self.get_table_schema(package)
        column_names = {x.name: x.display_name if x.display_name else x.name
                         for x in schema.column_schema}
        internal_columns = [x.name for x in schema.column_schema if x.internal]

        while True:
            resp = self._get_data_chunked(package, chunk_size=chunk_size, offset=offset, order_direction=order_direction, order_by=order_by)
            offset = offset + chunk_size

            df = pd.DataFrame.from_records(resp['rows'], exclude=internal_columns)
            df.columns = [column_names.get(c) for c in df.columns]

            yield df

            if len(df) < chunk_size:
                break

    @require_extension
    def get_tabular_data(self, package, limit, offset, order_by, order_direction):
        """
        Get data for tabular package using iterator
        """
        tab_iter = self.get_tabular_data_iter(package=package, offset=offset, order_by=order_by, order_direction=order_direction)
        df = pd.DataFrame()
        for tmp_df in tab_iter:
            df = df.append(tmp_df)
        return df[0:limit]

    def set_table_schema(self, package, tabular_schema):
        """
        Add a table schema to a tabular package.

        tabular_schema: blackfynn.tabular.models.TabularSchema
        """
        id = self._get_id(package)
        path = self._uri('/{id}/schema', id=id)
        if isinstance(tabular_schema,TabularSchema):
            body = { 'schema' : tabular_schema.as_dict()}
        else:
            body = { 'schema': tabular_schema}

        resp = self._post(path, json=body)
        data = resp

        return TabularSchema.from_dict(data)

    def get_table_schema(self, package):
        id = self._get_id(package)
        path = self._uri('/{id}/schema', id=id)
        resp = self._get(path)
        return TabularSchema.from_dict(resp)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Files
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class FilesAPI(APIBase):
    """
    Interface for managing file object in Blackfynn
    """
    base_uri = "/files"
    name = 'files'

    def create(self, file, destination=None):
        """
        Creates a file under the given destination or its current parent
        """
        container = file.parent if destination is None else destination

        body = file.as_dict()
        body["container"] = container

        response = self._post('', json=body)

        return File.from_dict(response, api=self.session)

    def update(self, file):
        """
        Update a file on the platform
        """
        loc = self._uri('/{id}', id=file.id)
        return self._put(loc, json=file.as_dict())

    def url(self, file):
        """
        Get pre-signed URL for File object
        """
        loc = self._uri('/{id}', id=file.id)
        return self._get(loc)['url']



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from future.utils import string_types

import itertools
import json
import requests

from blackfynn.api.base import APIBase
from blackfynn.models import (
    DataPackage,
    LinkedModelProperty,
    LinkedModelValue,
    Model,
    ModelProperty,
    ModelTemplate,
    ModelSelect,
    ModelFilter,
    ModelJoin,
    ProxyInstance,
    QueryResult,
    Record,
    RecordSet,
    Relationship,
    RelationshipProperty,
    RelationshipSet,
    RelationshipType
)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Models
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ModelsAPIBase(APIBase):

    def _get_concept_type(self, concept, instance = None):
        if isinstance(concept, Model):
            return concept.type
        elif isinstance(concept, string_types):
            return concept
        elif isinstance(instance, Record):
            return instance.type
        else:
            raise Exception("could not get concept type from concept {} or instance {} ".format(concept, instance))

    def _get_relationship_type(self, relationship, instance = None):
        if isinstance(relationship, RelationshipType):
            return relationship.type
        elif isinstance(relationship, string_types):
            return relationship
        elif isinstance(instance, Relationship):
            return instance.type
        else:
            raise Exception("could not get relationship type from relationship {} or instance {} ".format(relationship, instance))


class ModelsAPI(ModelsAPIBase):
    name = 'concepts'

    def __init__(self, session):
        self.host = session._host if session._model_service_host == None else session._model_service_host
        self.base_uri = "/models/datasets" if session._model_service_host == None else "/datasets"
        self.instances = RecordsAPI(session)
        self.relationships = ModelRelationshipsAPI(session)
        self.proxies = ModelProxiesAPI(session)
        self.query = ModelQueryAPI(session)
        super(ModelsAPI, self).__init__(session)

    def get_properties(self, dataset, concept):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        resp = self._get(self._uri('/{dataset_id}/concepts/{id}/properties', dataset_id=dataset_id, id=concept_id))
        return [ModelProperty.from_dict(r) for r in resp]

    def get_linked_properties(self, dataset, concept):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        resp = self._get(self._uri('/{dataset_id}/concepts/{id}/linked', dataset_id=dataset_id, id=concept_id))
        return {r['link']['name']: LinkedModelProperty.from_dict(r) for r in resp}

    def update_properties(self, dataset, concept):
        assert isinstance(concept, Model), "concept must be type Model"
        assert concept.schema, "concept schema cannot be empty"
        data = concept.as_dict()['schema']
        dataset_id = self._get_id(dataset)
        resp = self._put(self._uri('/{dataset_id}/concepts/{id}/properties', dataset_id=dataset_id, id=concept.id), json=data)
        return [ModelProperty.from_dict(r) for r in resp]

    def update_linked_property(self, dataset, concept, prop):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        prop_id = self._get_id(prop)
        resp = self._put(self._uri('/{dataset_id}/concepts/{id}/linked/{prop_id}', dataset_id=dataset_id, id=concept_id, prop_id=prop_id), json=prop.as_dict())
        return LinkedModelProperty.from_dict(resp)

    def delete_property(self, dataset, concept, prop):
        dataset_id  = self._get_id(dataset)
        concept_id  = self._get_id(concept)
        property_id = self._get_id(prop)
        return self._del(self._uri('/{dataset_id}/concepts/{concept_id}/properties/{property_id}',
                dataset_id  = dataset_id,
                concept_id  = concept_id,
                property_id = property_id))

    def delete_linked_property(self, dataset, concept, prop):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        prop_id = self._get_id(prop)
        self._del(self._uri('/{dataset_id}/concepts/{id}/linked/{prop_id}', dataset_id=dataset_id, id=concept_id, prop_id=prop_id))

    def get(self, dataset, concept):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        r = self._get(self._uri('/{dataset_id}/concepts/{id}', dataset_id=dataset_id, id=concept_id))
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        r['schema'] = self.get_properties(dataset, concept)
        r['linked'] = self.get_linked_properties(dataset, concept)
        return Model.from_dict(r, api=self.session)

    def delete(self, dataset, concept):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        return self._del(self._uri('/{dataset_id}/concepts/{id}', dataset_id=dataset_id, id=concept.id))

    def update(self, dataset, concept):
        assert isinstance(concept, Model), "concept must be type Model"
        data = concept.as_dict()
        data['id'] = concept.id
        dataset_id = self._get_id(dataset)
        r = self._put(self._uri('/{dataset_id}/concepts/{id}', dataset_id=dataset_id, id=concept.id), json=data)
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        if concept.schema:
            r['schema'] = self.update_properties(dataset, concept)
        if concept.linked:
            r['linked'] = {name: self.update_linked_property(dataset, concept, link) for name,link in concept.linked.items()}
        updated = Model.from_dict(r, api=self.session)
        return updated

    def create(self, dataset, concept):
        assert isinstance(concept, Model), "concept must be type Model"
        dataset_id = self._get_id(dataset)
        r = self._post(self._uri('/{dataset_id}/concepts', dataset_id=dataset_id), json=concept.as_dict())
        concept.id = r['id']
        r['dataset_id'] = r.get('dataset_id', dataset_id)

        if concept.schema:
            try:
                r['schema'] = self.update_properties(dataset, concept)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 400:
                    self._logger.error(
                        "Model was successfully created, but properties were "
                        "rejected with the following error: {}".format(str(
                            e.response.text.split('\n')[0]
                        ))
                    )
                else:
                    raise
        return Model.from_dict(r, api=self.session)

    def create_linked_property(self, dataset, concept, prop):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        assert prop.name not in self.get_linked_properties(dataset, concept), "Linked property '{}' already exists".format(prop.name)
        resp = self._post(self._uri('/{dataset_id}/concepts/{id}/linked', dataset_id=dataset_id, id=concept_id), json=prop.as_dict())
        return LinkedModelProperty.from_dict(resp)

    def create_linked_properties(self, dataset, concept, props):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        for p in props:
            assert p.name not in self.get_linked_properties(dataset, concept), "Linked property '{}' already exists".format(p.name)
        resp = self._post(self._uri('/{dataset_id}/concepts/{id}/linked/bulk', dataset_id=dataset_id, id=concept_id), json=[p.as_dict() for p in props])
        return [LinkedModelProperty.from_dict(r) for r in resp]

    def get_all(self, dataset):
        dataset_id = self._get_id(dataset)
        resp = self._get(self._uri('/{dataset_id}/concepts', dataset_id=dataset_id), stream=True)
        for r in resp:
            r['dataset_id'] = r.get('dataset_id', dataset_id)
            r['schema'] = self.get_properties(dataset, r['id'])
            r['linked'] = self.get_linked_properties(dataset, r['id'])
        concepts = [Model.from_dict(r, api=self.session) for r in resp]
        # for concept in concepts:
        #     concept.linked = {x.name: x for x in self.get_linked_properties(dataset, concept)}
        return { c.type: c for c in concepts }

    def delete_instances(self, dataset, concept, *instances):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        ids = [self._get_id(instance) for instance in instances]

        return self._del(self._uri('/{dataset_id}/concepts/{id}/instances', dataset_id=dataset_id, id=concept_id), json=ids)

    def files(self, dataset, concept, instance):
        """
        Return list of files (i.e. packages) related to record.
        """
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        instance_id = self._get_id(instance)
        resp = self._get(
            self._uri('/{dataset_id}/concepts/{concept_id}/instances/{instance_id}/files',
                dataset_id=dataset_id,
                concept_id=concept_id,
                instance_id=instance_id
            ))
        return [DataPackage.from_dict(pkg, api=self.session) for r,pkg in resp]

    def get_connected(self, dataset, model):
        """ Return a list of concepts related to the given model """
        dataset_id = self._get_id(dataset)
        model_id = self._get_id(model)
        resp = self._get(self._uri('/{dataset_id}/concepts/{model_id}/related',
                                   dataset_id=dataset_id,
                                   model_id=model_id),
                         stream=True)
        for r in resp:
            r['dataset_id'] = r.get('dataset_id', dataset_id)
            r['schema'] = self.get_properties(dataset, r['id'])
            r['linked'] = self.get_linked_properties(dataset, r['id'])

        concepts = [Model.from_dict(r, api=self.session) for r in resp]
        return {c.type: c for c in concepts}

    def get_related(self, dataset, concept):
        """ Return all SchemaRelationships and the Concepts they point to """
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        resp = self._get(
            self._uri('/{dataset_id}/concepts/{concept_id}/topology',
                      dataset_id=dataset_id, concept_id=concept_id))
        for r in resp:
            r['dataset_id'] = r.get('dataset_id', dataset_id)
            r['schema'] = self.get_properties(dataset, r['id'])
            r['linked'] = self.get_linked_properties(dataset, r['id'])
        concepts = [Model.from_dict(r, api=self.session) for r in resp]
        return concepts

    def get_topology(self, dataset):
        dataset_id = self._get_id(dataset)
        resp = self._get(
            self._uri('/{dataset_id}/concepts/schema/graph',
                      dataset_id=dataset_id))
        # What is returned is a list mixing
        results = {
            'models': [],
            'relationships': [],
            'linked_properties': []
        }
        for r in resp:
            r['dataset_id'] = r.get('dataset_id', dataset_id)
            if r.get('type')  == 'schemaRelationship':
                # This is a relationship
                results['relationships'].append(
                    Relationship.from_dict(r, api=self.session))
            elif r.get('type')  == 'schemaLinkedProperty':
                # This is a linked property type
                results['linked_properties'].append(
                    LinkedModelProperty.from_dict(r))
            else:
                # This is a model
                r['schema'] = self.get_properties(dataset, r['id'])
                r['linked'] = self.get_linked_properties(dataset, r['id'])
                results['models'].append(Model.from_dict(r, api=self.session))
        return results

    def get_summary(self, dataset):
        dataset_id = self._get_id(dataset)
        resp = self._get(
            self._uri('/{dataset_id}/concepts/graph/summary',
                      dataset_id=dataset_id))
        return resp

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model Instances
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class RecordsAPI(ModelsAPIBase):
    name = 'concepts.instances'

    def __init__(self, session):
        if session._model_service_host != None:
            self.host = session._model_service_host
        self.base_uri = "/models/datasets" if session._model_service_host == None else "/datasets"
        super(RecordsAPI, self).__init__(session)

    def get(self, dataset, instance, concept=None):
        dataset_id = self._get_id(dataset)
        instance_id = self._get_id(instance)
        concept_type = self._get_concept_type(concept, instance)

        r = self._get(self._uri('/{dataset_id}/concepts/{concept_type}/instances/{id}', dataset_id=dataset_id, concept_type=concept_type, id=instance_id))
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        return Record.from_dict(r, api=self.session)

    def relations(self, dataset, instance, related_concept, concept=None):
        dataset_id = self._get_id(dataset)
        instance_id = self._get_id(instance)
        related_concept_type = self._get_id(related_concept)
        concept_type = self._get_concept_type(concept, instance)

        res = self._get(self._uri('/{dataset_id}/concepts/{concept_type}/instances/{id}/relations/{related_concept_type}', dataset_id=dataset_id, concept_type=concept_type, id=instance_id, related_concept_type=related_concept_type))

        relations = []
        for r in res:
            relationship = r[0]
            concept = r[1]

            relationship['dataset_id'] = relationship.get('dataset_id', dataset_id)
            concept['dataset_id'] = concept.get('dataset_id', dataset_id)

            relationship = Relationship.from_dict(relationship, api=self.session)
            concept = Record.from_dict(concept, api=self.session)

            relations.append((relationship, concept))

        return relations

    def get_all(self, dataset, concept, limit=100, offset=0):
        dataset_id = self._get_id(dataset)
        concept_type = self._get_concept_type(concept)

        resp = self._get(self._uri('/{dataset_id}/concepts/{concept_type}/instances', dataset_id=dataset_id, concept_type=concept_type), params=dict(limit=limit, offset=offset), stream=True)
        for r in resp:
          r['dataset_id'] = r.get('dataset_id', dataset_id)
        instances = [Record.from_dict(r, api=self.session) for r in resp]

        return RecordSet(concept, instances)

    def delete(self, dataset, instance):
        assert isinstance(instance, Record), "instance must be type Record"
        dataset_id = self._get_id(dataset)
        return self._del(self._uri(
            '/{dataset_id}/concepts/{concept_type}/instances/{id}',
            dataset_id=dataset_id, concept_type=instance.type,
            id=instance.id
        ))

    def create(self, dataset, instance):
        assert isinstance(instance, Record), "instance must be type Record"
        dataset_id = self._get_id(dataset)
        r = self._post(self._uri('/{dataset_id}/concepts/{concept_type}/instances', dataset_id=dataset_id, concept_type=instance.type), json=instance.as_dict())
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        return Record.from_dict(r, api=self.session)

    def update(self, dataset, instance):
        assert isinstance(instance, Record), "instance must be type Record"
        dataset_id = self._get_id(dataset)
        r = self._put(self._uri('/{dataset_id}/concepts/{concept_type}/instances/{id}', dataset_id=dataset_id, concept_type=instance.type, id=instance.id), json=instance.as_dict())
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        return Record.from_dict(r, api=self.session)

    def create_many(self, dataset, concept, *instances):
        instance_type = instances[0].type
        for inst in instances:
            assert isinstance(inst, Record), "instance must be type Record"
            assert inst.type == instance_type, "Expected instance of type {}, found instance of type {}".format(instance_type, inst.type)
        dataset_id = self._get_id(dataset)
        values = [inst.as_dict() for inst in instances]
        resp = self._post(self._uri('/{dataset_id}/concepts/{concept_type}/instances/batch', dataset_id=dataset_id, concept_type=instance_type), json=values, stream=True)

        for r in resp:
            r['dataset_id'] = r.get('dataset_id', dataset_id)
        instances = [Record.from_dict(r, api=self.session) for r in resp]
        return RecordSet(concept, instances)

    def get_all_related(self, dataset, source_instance):
        related = self.get_counts(dataset, source_instance)
        return {
            item['name']: self.get_all_related_of_type(dataset, source_instance, item['name'])
            for item in related
            if not (item['name']=='package' and item['displayName']=='Files')
            # ^^ TODO: have API provide better means of distinguishing proxy vs. model
        }

    def get_all_related_of_type(self, dataset, source_instance, return_type, source_concept=None):
        """
        Return all records of type return_type related to instance.
        """
        dataset_id   = self._get_id(dataset)
        instance_id  = self._get_id(source_instance)
        instance_type = self._get_concept_type(source_concept, source_instance)

        resp = []
        limit = 100
        for offset in itertools.count(0, limit):
            batch = self._get(
                self._uri('/{dataset_id}/concepts/{instance_type}/instances/{instance_id}/relations/{return_type}',
                    dataset_id    = dataset_id,
                    instance_type = instance_type,
                    instance_id   = instance_id,
                    return_type   = return_type),
                params = {
                    'limit': limit,
                    'offset': offset})

            if not batch:
                break

            resp += batch

        for edge, node in resp:
            node['dataset_id'] = node.get('dataset_id', dataset_id)
        if not isinstance(return_type, Model):
            return_type = self.session.concepts.get(dataset, return_type)
        records = [Record.from_dict(r, api=self.session) for _,r in resp]
        return RecordSet(return_type, records)

    def get_counts(self, dataset, instance, concept=None):
        dataset_id   = self._get_id(dataset)
        concept_type = self._get_concept_type(concept, instance)
        instance_id  = self._get_id(instance)
        return self._get(self._uri('/{dataset_id}/concepts/{concept_type}/instances/{instance_id}/relationCounts',
                    dataset_id   = dataset_id,
                    concept_type = concept_type,
                    instance_id  = instance_id))

    def get_linked_values(self, dataset, concept, instance):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        instance_id = self._get_id(instance)
        resp = self._get(self._uri('/{dataset_id}/concepts/{id}/instances/{instance_id}/linked', dataset_id=dataset_id, id=concept_id, instance_id=instance_id))
        values = []
        for r in resp:
            link_type = concept.get_linked_property(r["schemaLinkedPropertyId"])
            target = concept._api.concepts.get(dataset, link_type.target)
            values.append(LinkedModelValue.from_dict(r, source_model=concept,
                target_model=target, link_type=link_type))
        return values

    def create_link(self, dataset, concept, instance, payload):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        instance_id = self._get_id(instance)

        # Delete any existing links of the given type:
        for link in self.get_linked_values(dataset, concept, instance):
            if link.type.id == payload['schemaLinkedPropertyId']:
                self.remove_link(dataset, concept, instance, link)

        resp = self._post(self._uri('/{dataset_id}/concepts/{id}/instances/{instance_id}/linked', dataset_id=dataset_id, id=concept_id, instance_id=instance_id), json=payload)
        link_type = concept.get_linked_property(resp["schemaLinkedPropertyId"])
        target = concept._api.concepts.get(dataset, link_type.target)
        return LinkedModelValue.from_dict(resp, source_model=concept,
            target_model=target, link_type=link_type)

    def remove_link(self, dataset, concept, instance, value):
        dataset_id = self._get_id(dataset)
        concept_id = self._get_id(concept)
        instance_id = self._get_id(instance)
        link_id = self._get_id(value)
        self._del(self._uri('/{dataset_id}/concepts/{id}/instances/{instance_id}/linked/{link_id}', dataset_id=dataset_id, id=concept_id, instance_id=instance_id, link_id=link_id))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Relationships
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ModelRelationshipsAPI(ModelsAPIBase):
    name = 'concepts.relationships'

    def __init__(self, session):
        if session._model_service_host != None:
            self.host = session._model_service_host
        self.base_uri = "/models/datasets" if session._model_service_host == None else "/datasets"
        self.instances = ModelRelationshipInstancesAPI(session)
        super(ModelRelationshipsAPI, self).__init__(session)

    def create(self, dataset, relationship):
        assert isinstance(relationship, RelationshipType), "Must be of type Relationship"
        dataset_id = self._get_id(dataset)
        rel_dict = relationship.as_dict()
        r = self._post(self._uri('/{dataset_id}/relationships', dataset_id=dataset_id), json=rel_dict)
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        return RelationshipType.from_dict(r, api=self.session)

    def get(self, dataset, relationship):
        dataset_id = self._get_id(dataset)
        relationship_id = self._get_id(relationship)
        r = self._get(self._uri('/{dataset_id}/relationships/{r_id}', dataset_id=dataset_id, r_id=relationship_id))
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        return RelationshipType.from_dict(r, api=self.session)

    def get_all(self, dataset):
        dataset_id = self._get_id(dataset)
        resp = self._get(self._uri('/{dataset_id}/relationships', dataset_id=dataset_id), stream=True)
        for r in resp:
          r['dataset_id'] = r.get('dataset_id', dataset_id)
        relations = [RelationshipType.from_dict(r, api=self.session) for r in resp]
        return {r.type: r for r in relations}

class ModelRelationshipInstancesAPI(ModelsAPIBase):
    name = 'concepts.relationships.instances'

    def __init__(self, session):
        if session._model_service_host != None:
            self.host = session._model_service_host
        self.base_uri = "/models/datasets" if session._model_service_host == None else "/datasets"
        super(ModelRelationshipInstancesAPI, self).__init__(session)

    def get_all(self, dataset, relationship):
        dataset_id = self._get_id(dataset)
        relationship_id = self._get_id(relationship)

        resp = self._get(self._uri('/{dataset_id}/relationships/{r_id}/instances', dataset_id=dataset_id, r_id=relationship_id), stream=True)
        for r in resp:
          r['dataset_id'] = r.get('dataset_id', dataset_id)
        instances = [Relationship.from_dict(r, api=self.session) for r in resp]
        return RelationshipSet(relationship, instances)

    def get(self, dataset, instance, relationship=None):
        dataset_id = self._get_id(dataset)
        instance_id = self._get_id(instance)
        relationship_type = self._get_relationship_type(relationship, instance)
        r = self._get( self._uri('/{dataset_id}/relationships/{r_type}/instances/{id}', dataset_id=dataset_id, r_type=relationship_type, id=instance_id))
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        return Relationship.from_dict(r, api=self.session)

    def delete(self, dataset, instance):
        assert isinstance(instance, Relationship), "instance must be of type Relationship"
        dataset_id = self._get_id(dataset)
        return self._del( self._uri('/{dataset_id}/relationships/{r_type}/instances/{id}', dataset_id=dataset_id, r_type=instance.type, id=instance.id))

    def link(self, dataset, relationship, source, destination, values=dict()):
        assert isinstance(source, (Record, DataPackage)), "source must be an object of type Record or DataPackage"
        assert isinstance(destination, (Record, DataPackage)), "destination must be an object of type Record or DataPackage"

        if isinstance(source, DataPackage):
            assert isinstance(destination, Record), "DataPackages can only be linked to Records"
            return self.session.concepts.proxies.create(dataset, source.id, relationship, destination, values, "ToTarget", "package")
        elif isinstance(destination, DataPackage):
            assert isinstance(source, Record), "DataPackages can only be linked to Records"
            return self.session.concepts.proxies.create(dataset, destination.id, relationship, source, values, "FromTarget", "package")
        else:
            dataset_id = self._get_id(dataset)
            relationship_type = self._get_relationship_type(relationship)
            values = [dict(name=k, value=v) for k,v in values.items()]
            instance = Relationship(dataset_id=dataset_id, type=relationship_type, source=source, destination=destination, values=values)
            return self.create(dataset, instance)

    def create(self, dataset, instance):
        assert isinstance(instance, Relationship), "instance must be of type Relationship"
        dataset_id = self._get_id(dataset)
        resp = self._post( self._uri('/{dataset_id}/relationships/{r_type}/instances', dataset_id=dataset_id, r_type=instance.type), json=instance.as_dict())
        r = resp[0] # responds with list
        r['dataset_id'] = r.get('dataset_id', dataset_id)
        return Relationship.from_dict(r, api=self.session)

    def create_many(self, dataset, relationship, *instances):
        assert all([isinstance(i, Relationship) for i in instances]), "instances must be of type Relationship"
        instance_type = instances[0].type
        dataset_id = self._get_id(dataset)
        values = [inst.as_dict() for inst in instances]
        resp = self._post( self._uri('/{dataset_id}/relationships/{r_type}/instances/batch', dataset_id=dataset_id, r_type=instance_type), json=values)

        for r in resp:
            r[0]['dataset_id'] = r[0].get('dataset_id', dataset_id)
        instances =  [Relationship.from_dict(r[0], api=self.session) for r in resp]
        return RelationshipSet(relationship, instances)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model Proxies
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ModelProxiesAPI(ModelsAPIBase):
    name = 'concepts.proxies'

    proxy_types = ["package"]
    direction_types = ["FromTarget", "ToTarget"]

    def __init__(self, session):
        if session._model_service_host != None:
            self.host = session._model_service_host
        self.base_uri = "/models/datasets" if session._model_service_host == None else "/datasets"
        super(ModelProxiesAPI, self).__init__(session)

    def create(self, dataset, external_id, relationship, concept_instance, values, direction = "ToTarget", proxy_type = "package", concept = None):
        assert proxy_type in self.proxy_types, "proxy_type must be one of {}".format(self.proxy_types)
        assert direction in self.direction_types, "direction must be one of {}".format(self.direction_types)

        dataset_id = self._get_id(dataset)
        concept_instance_id = self._get_id(concept_instance)
        concept_type = self._get_concept_type(concept, concept_instance)
        relationship_type = self._get_relationship_type(relationship)
        relationshipData = [dict(name=k, value=v) for k,v in values.items()]

        request = {}
        request['externalId'] = external_id
        request['conceptType'] = concept_type
        request['conceptInstanceId'] = concept_instance_id
        request['targets'] = [
            {
                'direction': direction,
                'linkTarget': {
                    "ConceptInstance": {
                        'id': concept_instance_id
                    }
                },
                'relationshipType': relationship_type,
                'relationshipData': relationshipData
            }
        ]

        r = self._post(self._uri('/{dataset_id}/proxy/{p_type}/instances',
                                 dataset_id=dataset_id, p_type=proxy_type),
                       json=request)
        instance = r[0]['relationshipInstance']
        instance['dataset_id'] = instance.get('dataset_id', dataset_id)
        return Relationship.from_dict(instance, api=self.session)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model Query
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ModelQuery(object):
    allowed_operators = set(['eq', 'neq', 'lt', 'lte', 'gt', 'gte'])

    def __init__(self, query_api, model, dataset_id):
        self.query_api = query_api
        self.model = model
        self.dataset_id = dataset_id
        self._select = None
        self._filters = []
        self._joins = []
        self._offset = 0
        self._limit = 50

    def select(self, *join_keys):
        """
        Add a select clause to the query.

        Args:
            join_keys: (List[str|model]) The names of the targets to select on.

        Examples::

            Review.query() \
                .filter("is_complete", "eq", False) \
                .join("reviewer", ("id", "eq", "12345")) \
                .select(reviewer")
                .run()
        """
        self._select = ModelSelect(*join_keys)
        return self

    def filter(self, key, operator, value):
        """
        Add a filter to the query.

        Args:
            key: (string) The name of the property to test.

            operator: (string) The predicate operator.

            value (any) The right hand of the predicate to match against.

        Returns:
            The current query object.

        Example::

            Review.query() \
                .filter("is_complete", "eq", False) \
                .run()
        """
        assert isinstance(key, str), "key must be a string"
        assert operator in self.allowed_operators, "not a valid predicate operator: {}".format(operator)
        self._filters.append(ModelFilter(key, operator, value))
        return self

    def join(self, target, *filters):
        """
        Add a join clause to the query.

        Args:
            target: (string|model) The model to join to via some relationship.

            filters: (*tuple) Filter conditions to apply to the join. Each
                tuple is expected to be a triple (key, operator, value), much
                like the constructor for `ModelFilter`.

        Returns:
            The current query object.

        Example::

            Review.query() \
                .filter("is_complete", "eq", False) \
                .join("reviewer", ("id", "eq", "12345")) \
                .select(reviewer")
                .run()
        """
        assert isinstance(target, (str, Model)), "target must be a string or a model"
        self._joins.append(ModelJoin(target, *filters))
        return self

    def offset(self, value):
        """
        Adds an offset to the query.

        Args:
            value (int) The offset value.

        Returns:
            The current query object.

        Example::

            Review.query() \
                .filter("is_complete", "eq", False) \
                .offset(100)
                .run()
        """
        self._offset = value
        return self

    def limit(self, value):
        """
        Adds an limit to the query.

        Args:
            value (int) The limit value.

        Returns:
            The current query object.

        Example::

            Review.query() \
                .filter("is_complete", "eq", False) \
                .limit(100)
                .run()
        """
        self._limit = value
        return self

    def _build_query(self):
        query = {
            "type": { "concept": { "type": self.model.type } },
            "filters": [f.as_dict() for f in self._filters],
            "joins": [j.as_dict() for j in self._joins],
            "offset": self._offset,
            "limit": self._limit,
            "orderBy": { "Ascending": { "field": "$createdAt" } },
        }
        if self._select is not None:
            query['select'] = self._select.as_dict()
        return query

    def run(self):
        """
        Run the constructed query.

        Returns:
            A list of matching Record instances.
        """
        resp = self.query_api._post(self.query_api._uri('/{dataset_id}/query/run', dataset_id=self.dataset_id), json=self._build_query())
        if resp is None:
            return []

        records = []
        for r in resp:
            # get the target first:
            target_value = r['targetValue']
            target_value['dataset_id'] = self.dataset_id
            target = Record.from_dict(target_value, api=self.query_api.session)

            # then any attached records by join type:
            joined = {}
            if self._select is not None:
                for join_key in self._select.join_keys:
                    if join_key in r:
                        join_value = r[join_key]
                        join_value['dataset_id'] = self.dataset_id
                        joined[join_key] = Record.from_dict(join_value, api=self.query_api.session)

            records.append(QueryResult(self.dataset_id, target, joined))

        return records

class ModelQueryAPI(ModelsAPIBase):
    name = 'concepts.query'

    def __init__(self, session):
        if session._model_service_host != None:
            self.host = session._model_service_host

        self.base_uri = "/models/datasets" if session._model_service_host == None else "/datasets"
        super(ModelQueryAPI, self).__init__(session)

    def new(self, model, dataset_id):
        """
        Construct a new query.

        Args:
            model: The model to use as the join target.

            dataset_id: The ID of the dataset the model is contained in.

        Returns:
            A list of records fulfilling the conditions of the query.

        Example::

            self.new(model, dataset_id) \
                .filter("is_completed", "eq", False) \
                .run()
        """
        return ModelQuery(self, model, dataset_id)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Model Templates
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ModelTemplatesAPI(APIBase):

    base_uri = "/model-schema"
    name = 'templates'

    def get_all(self):
        org_id = self._get_int_id(self.session._context)
        resp = self._get(self._uri('/organizations/{orgId}/templates', orgId=org_id))
        return [ModelTemplate.from_dict(t) for t in resp]

    def get(self, template_id):
        org_id = self._get_int_id(self.session._context)
        resp = self._get(self._uri('/organizations/{orgId}/templates/{templateId}',
                                   orgId=org_id, templateId=template_id))
        return ModelTemplate.from_dict(resp)

    def validate(self, template):
        assert isinstance(template, ModelTemplate), "template must be type Template"
        response = self._post(self._uri('/validate'), json=template.as_dict())
        if response == "":
            return True
        else:
            return response

    def create(self, template):
        assert isinstance(template, ModelTemplate), "template must be type Template"
        org_id = self._get_int_id(self.session._context)
        resp = self._post(self._uri('/organizations/{orgId}/templates', orgId=org_id), json=template.as_dict())
        return ModelTemplate.from_dict(resp)

    def apply(self, dataset, template):
        org_id = self._get_int_id(self.session._context)
        dataset_id = self._get_int_id(dataset)
        template_id = self._get_id(template)
        resp = self._post(self._uri('/organizations/{orgId}/templates/{templateId}/datasets/{datasetId}',
                                    orgId=org_id, templateId=template_id, datasetId=dataset_id))
        return [ModelProperty.from_dict(t) for t in resp]

    def delete(self, template_id):
        org_id = self._get_int_id(self.session._context)
        return self._del(self._uri('/organizations/{orgId}/templates/{templateId}',
                                   orgId=org_id, templateId=template_id))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function
from future.utils import string_types, integer_types

import datetime

from blackfynn.extensions import require_extension, pandas as pd, numpy as np

# data type helpers

def value_as_type(value, dtype):
    try:
        if dtype == 'string':
            return str(value)
        elif dtype == 'integer':
            return int(value)
        elif dtype == 'double':
            return float(value)
        elif dtype == 'date':
            return infer_epoch_msecs(value)
        if dtype == 'boolean':
            return value.lower()=='true'
    except:
        raise Exception('Unable to set value={} as type {}'.format(value, dtype))


def get_data_type(v):
    """
    Infers type from value. Returns tuple of (type, value)
    """
    if isinstance(v, datetime.datetime):
        return ("date", msecs_since_epoch(v))
    elif isinstance(v, bool):
        return ("boolean", str(v).lower())
    elif isinstance(v, float):
        return ("double", v)
    elif isinstance(v, integer_types):
        return ("integer", v)
    else:
        # infer via casting
        if is_integer(v):
            return ("integer", int(v))
        elif is_decimal(v):
            return ("double", float(v))
        else:
            return ("string", str(v))

def is_integer(s):
    try:
        int(s)
        return True
    except ValueError:
        return False

def is_decimal(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

# time-series helpers

def infer_epoch_msecs(thing):
    if isinstance(thing, datetime.datetime):
        return msecs_since_epoch(thing)
    elif isinstance(thing, (integer_types, float)):
        # assume milliseconds
        return int(thing)
    elif isinstance(thing, string_types):
        # attempt to convert to msec integer
        return int(thing)
    else:
        raise Exception("Cannot parse date")

def infer_epoch(thing):
    if isinstance(thing, datetime.datetime):
        return usecs_since_epoch(thing)
    elif isinstance(thing, (integer_types, float)):
        # assume microseconds
        return int(thing)
    else:
        raise Exception("Cannot parse date")

def secs_since_epoch(the_time):
    the_time = the_time.replace(tzinfo=None)
    # seconds from epoch (float)
    return (the_time-datetime.datetime.utcfromtimestamp(0)).total_seconds()

def msecs_since_epoch(the_time):
    # milliseconds from epoch (integer)
    return int(secs_since_epoch(the_time)*1000)

def usecs_since_epoch(the_time):
    # microseconds from epoch (integer)
    return int(secs_since_epoch(the_time)*1e6)

def usecs_to_datetime(us):
    # convert usecs since epoch to proper datetime object
    return datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(microseconds=int(us))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Timeseries helpers
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

@require_extension
def generate_data(size, func='walk', scale=5, periods=10):

    remainder = size%periods
    pattern_length = int(size/periods)
    if func=='walk':
        x = (np.random.rand(size)-0.5).cumsum()
        # normalize to [-scale,scale]
        return x/np.max([x.min(), x.max()])*scale
    elif func=='sin':
        return np.sin(np.linspace(0, np.pi*periods, size))*scale
    elif func=='square':
        pattern = np.concatenate([np.ones(pattern_length//2)*-1, np.ones(pattern_length//2)])
        return np.concatenate([np.repeat(pattern, periods), pattern[:remainder]])*scale
    elif func=='sawtooth':
        pattern = np.linspace(-scale,scale,pattern_length)
        return np.concatenate([np.repeat(pattern, periods),pattern[:remainder]])


@require_extension
def generate_dataframe(minutes=2, freq=100):

    start = datetime.datetime(2017, 1, 1, 0, 0)
    end   = start + datetime.timedelta(minutes=minutes)

    # index
    ind = pd.date_range(start=start, end=end, freq='10000u', closed='left')
    n = len(ind)

    # dataframe
    return pd.DataFrame({
        'random-walk': generate_data(n, func='walk'),
        'sin': generate_data(n, func='sin'),
        'sawtooth': generate_data(n, func='sawtooth'),
        'square': generate_data(n, func='square')
    }, index=ind)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function
from builtins import dict, object
from future.utils import string_types

import io
import os
import sys
import threading
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed

import boto3
from boto3.s3.transfer import S3Transfer
from botocore.client import Config

import blackfynn.log as log
from blackfynn.api.agent import validate_agent_installation, agent_upload
from blackfynn.api.base import APIBase
from blackfynn.models import Collection, DataPackage, Dataset, TimeSeries

# GLOBAL
UPLOADS = {}

logger = log.get_logger('blackfynn.api.transfers')


def check_files(files):
    for f in files:
        if not os.path.exists(f):
            raise Exception("File {} not found.".format(f))


class ProgressPercentage(object):
    def __init__(self, filename, upload_session_id):
        self._filename = filename
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()
        self._done = (self.progress == 1)
        self._errored = False

        if upload_session_id is None:
            upload_session_id = str(uuid.uuid4())
        self._session_id = upload_session_id

    def __call__(self, bytes_amount):
        global UPLOADS
        # To simplify we'll assume this is hooked up
        # to a single filename.
        if self._errored:
            return
        with self._lock:
            self._seen_so_far += bytes_amount
            self._done = self._seen_so_far >= self._size
            UPLOADS[self._session_id] = self

    @property
    def progress(self):
        if self._size == 0:
            return 1
        return (self._seen_so_far / self._size)

    def set_error(self):
        self._errored = True


class UploadManager(object):
    def __init__(self):
        self.id = str(uuid.uuid4())
        # map of filename -> progress callback
        self._uploads = {}
        self._exit = False

    def init_file(self, filename):
        global UPLOADS
        # filename -> session
        uid = str(uuid.uuid4())
        self._uploads[filename] = uid
        # init to blank progress
        UPLOADS[uid] = ProgressPercentage(filename, None)
        return uid

    @property
    def uploads(self):
        return [(f, UPLOADS[uid]) for f, uid in self._uploads.items()]

    @property
    def done(self):
        return sum([fstat._done for _,fstat in self.uploads]) == len(self._uploads)

    def _print_progress(self, width=16):
        for f, fstat in self.uploads:

            if fstat._done:
                state = 'DONE'
            elif fstat._errored:
                state = 'ERRORED'
                self._exit = True
            elif fstat._seen_so_far>0:
                state = 'UPLOADING'
            else:
                state = 'WAITING'

            text = ' [ {bars}{dashes} ] {state:12s} {percent:05.1f}% {name}\n'.format(
                        bars = '#'*int(fstat.progress*width),
                        dashes = '-'*(width - int(fstat.progress*width)),
                        percent = fstat.progress*100,
                        name = os.path.basename(f),
                        state = state)

            sys.stdout.write('{}\r'.format(text))
            sys.stdout.flush()

        # move cursor to relative beginning
        sys.stdout.write("\033[F"*len(self.uploads))

    def display_progress(self):
        while True:
            self._print_progress()
            time.sleep(0.2)
            if self.done or self._exit: break
        self._print_progress()
        # move cursor to bottom
        sys.stdout.write('\n'*len(self.uploads))


def upload_file(
        file,
        s3_host,
        s3_port,
        s3_bucket,
        s3_keybase,
        region,
        access_key_id,
        secret_access_key,
        session_token,
        encryption_key_id,
        upload_session_id=None,
        ):

    # progress callback
    progress = ProgressPercentage(file, upload_session_id)
    UPLOADS[upload_session_id] = progress

    try:
        # account for dev connections
        resource_args = {}
        config_args = dict(signature_version='s3v4')
        if 'amazon' not in s3_host.lower() and len(s3_host)!=0:
            resource_args = dict(endpoint_url="http://{}:{}".format(s3_host, s3_port))
            config_args = dict(s3=dict(addressing_style='path'))

        # connect to s3
        session = boto3.session.Session()
        s3 = session.client('s3',
            region_name = region,
            aws_access_key_id = access_key_id,
            aws_secret_access_key = secret_access_key,
            aws_session_token = session_token,
            config = Config(**config_args),
            **resource_args
        )

        # s3 key
        s3_key = '{}/{}'.format(s3_keybase, os.path.basename(file))

        # upload file to s3
        s3.upload_file(
            Filename=file,
            Bucket=s3_bucket,
            Key=s3_key,
            Callback=progress,
            ExtraArgs=dict(
                ServerSideEncryption="aws:kms",
                SSEKMSKeyId=encryption_key_id
            ))

        return s3_key

    except Exception as e:
        logger.debug(e)
        progress.set_error()
        raise e


class IOAPI(APIBase):
    """
    Input/Output interface.
    """
    name = 'io'

    def upload_files(self, destination, files, dataset=None, append=False,
                     display_progress=False, recursive=False, use_agent=True):
        if isinstance(destination, Dataset):
            # uploading into dataset
            destination_id = None
            dataset = destination
            dataset_id = self._get_id(dataset)
        elif destination is None and dataset is not None:
            # uploading into dataset
            destination_id = None
            dataset_id = self._get_id(dataset)
        elif isinstance(destination, Collection):
            # uploading into collection
            destination_id = self._get_id(destination)
            dataset_id = self._get_id(destination.dataset)
        elif append and isinstance(destination, TimeSeries):
            # uploading into timeseries package must be an append
            destination_id = self._get_id(destination)
            dataset_id = self._get_id(destination.dataset)
        elif isinstance(destination, string_types):
            # assume ID is for collection
            if dataset is None:
                raise Exception("Must also supply dataset when specifying destination by ID")
            destination_id = destination
            dataset_id = self.get_id(dataset)
        else:
            raise Exception("Cannot upload to destination of type {}".format(type(destination)))

        # check input files
        check_files(files)

        # sanity check dataset
        try:
            if isinstance(dataset, Dataset) and dataset.exists:
                ds = dataset
            else:
                ds = self.session.datasets.get(dataset_id)
        except:
            raise Exception("dataset does not exist")

        if destination_id is not None:
            try:
                if isinstance(destination, (Dataset, Collection, DataPackage)) and destination.exists:
                    pass
                else:
                    destination = self.session.packages.get(destination_id)
            except:
                raise Exception("destination does not exist")

            # check type for appends
            if append:
                if not isinstance(destination, TimeSeries):
                    raise Exception("Append destination must be TimeSeries package.")
            else:
                if not isinstance(destination, Collection):
                    raise Exception("Upload destination must be Collection or Dataset.")

        if use_agent:
            # Push uploads through the Blackfynn agent
            validate_agent_installation(self.session.settings)
            return agent_upload(
                destination=destination,
                files=files,
                dataset=ds,
                append=append,
                recursive=recursive,
                display_progress=display_progress,
                settings=self.session.settings)

        elif any(os.path.isdir(f) for f in files):
            raise Exception(
                "Blackfynn Agent required to upload directories\n"
                "Visit https://developer.blackfynn.io/agent for installation directions and pass `use_agent=True` to `upload`.")

        elif recursive:
            raise Exception(
                "Blackfynn Agent required to upload recursively\n"
                "Visit https://developer.blackfynn.io/agent for installation directions and pass `use_agent=True` to `upload`.`")


        # get preview
        import_id_map = self.get_preview(files, append)

        # get upload credentials
        resp = self.session.security.get_upload_credentials(dataset_id)
        creds = resp['tempCredentials']
        s3_bucket = resp['s3Bucket']
        region = creds['region']
        access_key_id = creds['accessKey']
        secret_access_key = creds['secretKey']
        session_token = creds['sessionToken']
        encryption_key_id = resp['encryptionKeyId']

        # upload session manager
        upload_session = UploadManager()

        # parallel upload to s3
        file_results = {}
        group_results = []
        with ThreadPoolExecutor(max_workers=min(len(files),self.session.settings.max_upload_workers)) as e:
            futures = {
                e.submit(
                    fn = upload_file,
                    file = file,
                    s3_host=self.session.settings.s3_host,
                    s3_port=self.session.settings.s3_port,
                    s3_bucket = s3_bucket,
                    s3_keybase = '{}/{}'.format(resp['s3Key'], import_id_map[file]),
                    region = region,
                    access_key_id = access_key_id,
                    secret_access_key = secret_access_key,
                    session_token = session_token,
                    encryption_key_id = encryption_key_id,
                    upload_session_id = upload_session.init_file(file),
                ): file
                for file in files
            }
            # thread for displaying progress
            if display_progress:
                e.submit(upload_session.display_progress())
            # retrieve results
            for future in as_completed(futures):
                fname = futures[future]
                file_results[fname] = future.result()
                if future.exception() is not None:
                    raise future.exception()
                import_id = import_id_map[fname]
                # check to see if rest of the import group has uploaded
                if all([ name in file_results for name, id in import_id_map.items() if id == import_id ]):
                    # trigger ETL import
                    group_results.append(self.set_upload_complete(import_id, dataset_id, destination_id, append))

        return group_results

    def get_preview(self, files, append):
        params = dict(
            append = append,
        )

        payload = { "files": [
            {
                "fileName": os.path.basename(f),
                "size": os.path.getsize(f),
                "uploadId": i,
            } for i, f in enumerate(files)
        ]}

        response = self._post(
            endpoint = self._uri('/files/upload/preview'),
            params = params,
            json=payload,
            )

        import_id_map = dict()
        for p in response.get("packages", list()):
            import_id = p.get("importId")
            warnings = p.get("warnings", list())
            for warning in warnings:
                logger.warn("API warning: {}".format(warning))
            for f in p.get("files", list()):
                index = f.get("uploadId")
                import_id_map[files[index]] = import_id
        return import_id_map

    def set_upload_complete(self, import_id, dataset_id, destination_id, append=False, targets=None):
        params = dict(
            append = append,
            datasetId = dataset_id,
            importId = import_id,
        )
        if destination_id is not None:
            params['destinationId'] = destination_id

        return self._post(
            endpoint=self._uri('/files/upload/complete/{import_id}', import_id=import_id),
            params=params,
            json = targets
            )



import pytest

from blackfynn.models import ModelProperty, ModelTemplate


def test_templates(client, dataset):

    template_1 = ModelTemplate(
        name="Test Template",
        description="This is a test",
        category="Person",
        properties=dict(
            name=dict(
                type="string",
                description="Name"
            ),
            DOB=dict(
                type="string",
                format="date",
                description="Date of Birth"
            ),
            Weight=dict(
                type="number",
                unit="lbs",
                description="Weight"
            )
        ),
        required=["name"]
    )

    # Validate template schema
    validation_response_1 = client.validate_model_template(template=template_1)

    assert(validation_response_1 == True)

    validation_response_2 = client.validate_model_template(
        name="Validation Test",
        category="Person",
        properties=[("name", "string"), ("Weight", "number")]
    )

    assert (validation_response_2 == True)

    with pytest.raises(Exception):
        client.validate_model_template(
        name="Invalid Template",
        category="Test",
        properties=[("name", "string"), ("Weight", "invalid-type")]
    )

    # Create a new template
    new_template = client.create_model_template(template=template_1)
    assert(isinstance(new_template, ModelTemplate))

    # Create another new template
    template_2 = client.create_model_template(
        name="Another Test Template",
        category="Person",
        properties=[("name", "string"), ("DOB", "date"), ("Weight", "number")]
    )
    assert (isinstance(template_2, ModelTemplate))

    # Try to create an invalid template
    with pytest.raises(Exception):
        client.create_model_template(
            name="Bad Template",
            category="Person",
            properties=[("invalid:name", "string")]
        )

    # Get a template
    result = client.get_model_template(new_template.id)
    assert (isinstance(result, ModelTemplate))
    assert(result.description == "This is a test")
    assert(result.category == "Person")

    # Get all templates belonging to the current organization
    all_templates = client.get_model_templates()
    assert (isinstance(all_templates[0], ModelTemplate))

    # Apply a template to a dataset
    applied = dataset.import_model(new_template)
    assert (isinstance(applied[0], ModelProperty))

    # Delete templates
    client.delete_model_template(new_template.id)
    client.delete_model_template(template_2.id)

    with pytest.raises(Exception):
        client.get_model_template(new_template.id)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function
from builtins import dict, object, range, zip
from future.utils import as_native_str, integer_types, string_types

import datetime
import itertools
import math
import re

from concurrent.futures import ThreadPoolExecutor
from itertools import count, islice
import requests

# blackfynn
from blackfynn.api.base import APIBase
from blackfynn.cache import get_cache
from blackfynn.extensions import require_extension, pandas as pd, numpy as np
from blackfynn.models import (
    File,
    TimeSeries,
    TimeSeriesAnnotation,
    TimeSeriesAnnotationLayer,
    TimeSeriesChannel,
    get_package_class
)
from blackfynn.utils import infer_epoch, usecs_since_epoch, usecs_to_datetime

cache = None

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Helpers
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def parse_timedelta(time):
    """
    Returns microseconds of time expression, where time can be of the forms:
     - string:  e.g. '1s', '5m', '3h'
     - delta:   datetime.timedelta object
    """
    if isinstance(time, string_types):
        # parse string into timedelta
        regex = re.compile(r'((?P<hours>\d*\.*\d+?)hr)?((?P<minutes>\d*\.*\d+?)m)?((?P<seconds>\d*\.*\d+?)s)?')
        parts = regex.match(time)
        if not parts:
            return
        parts = parts.groupdict()
        time_params = {}
        for (name, param) in parts.items():
            if param:
                time_params[name] = float(param)
        time = datetime.timedelta(**time_params)

    if isinstance(time, datetime.timedelta):
        # return microseconds
        return int(time.total_seconds()*1e6)

    elif isinstance(time, (integer_types, float)):
        # assume already in microseconds
        return time

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# TimeSeries Request
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class ChannelPage(object):
    def __init__(self, channel, page, settings, use_cache=True):
        self.channel   = channel
        self.page      = int(page)
        self.use_cache = use_cache

        page_size = settings.ts_page_size
        global cache
        if self.use_cache and cache is None:
            cache = get_cache(settings, start_compaction=True)
            page_size = cache.page_size

        # fixed page -- determined from epoch(0)
        pg_delta = channel._page_delta(page_size)
        self.start = int(self.page  * pg_delta)
        self.stop  = int(self.start + pg_delta)

        # current response
        self.data  = None

    def get(self, api):
        cache_exists = False
        update_cache = False

        # check if page is cached
        if self.use_cache:
            cache_exists = cache.check_page(self.channel, self.page)
            if cache_exists:
                # we (should) have cache, try to use existing cache entry
                self.data = cache.get_page_data(self.channel, self.page)
                if self.data is None:
                    # cache entry has disappeared, let's update it
                    update_cache = True
                else:
                    return self.data

        # make request
        args = dict(
            # Note: uses streaming server
            host     = api._host,
            endpoint = '/streaming/ts/retrieve/continuous',
            base     = '',
            params   = dict(
                channel = self.channel.id,
                limit   = '', # required by API
                session = api.headers.get('X-SESSION-ID'),
                start   = self.start,
                end     = self.stop)
        )
        data = api._get(**args)
        self.data = self._load_data(data)

        # save page to cache
        if self.use_cache and (not cache_exists or update_cache):
            cache.set_page_data(self.channel, self.page, self.data, update=update_cache)

        return self.data

    @require_extension
    def _load_data(self, data, datetime_index=True):
        vec_usecs_to_datetime = np.vectorize(usecs_to_datetime)

        # handle data response
        times = np.array( [t[0] for t in data] )
        data  = np.array( [d[1] for d in data] )
        # fix -- sometimes API responds out-of-order
        order = np.argsort(times)
        times = times[order]
        data  = data[order]

        if datetime_index and len(times)>0:
            times = vec_usecs_to_datetime(times)

        # return pandas series
        return pd.Series( data=data, index=times, name=str(self.channel))


class ChannelIterator(object):
    """
    We make requests to API/cache using some fixed page-size, but the
    user typically wants data results in some specified "chunk size".
    This accumulates the data pages in order to serve the data back
    in the specified chunk size.
    """
    def __init__(self, channel, start, stop, chunk_time, api, use_cache=True):
        self.channel    = channel
        self.start      = start
        self.stop       = stop
        self.start_dt   = usecs_to_datetime(self.start)
        self.stop_dt    = usecs_to_datetime(self.stop)
        self.use_cache  = use_cache
        self.api        = api

        # page delta (usecs) for channel
        self.page_delta = channel._page_delta(api.settings.ts_page_size)

        # page iteration
        self.page_start = int(math.floor(self.start/(1.0*self.page_delta)))
        self.page_end   = int(math.ceil(self.stop /(1.0*self.page_delta)))

        # chunk iteration
        self.chunk_per_page = chunk_time is None
        # chunk over specfied time
        if not self.chunk_per_page:
            self.chunk_time = int(chunk_time) # in usecs
            self.chunk_size = int(channel.rate * self.chunk_time/1.0e6)
        self.chunk  = None
        self.offset = usecs_to_datetime(self.start)

    @require_extension
    def get_chunks(self):
        # page size may be more/less than requested data

        self.chunk = pd.Series()
        pages = iter(np.arange(self.page_start, self.page_end))
        page = None
        while True:
            if self.chunk_per_page or \
               (page is None and len(self.chunk) < self.chunk_size):
                # get next page
                try: p = next(pages)
                except: break
                page = ChannelPage(
                        settings = self.api.settings,
                        channel   = self.channel,
                        page      = p,
                        use_cache = self.use_cache)
                data = page.get(self.api)
                # no more data
                if data is None: break
                # grow data series
                i_start = None
                i_stop  = None
                if page.start < self.start:
                    i_start = self.start_dt
                if page.stop >= self.stop:
                    i_stop = self.stop_dt
                else:
                    # more pages needed - reset
                    page = None
                data_slice = data.loc[i_start:i_stop]
                if self.chunk_per_page:
                    yield data_slice
                else:
                    self.chunk = self.chunk.append(data_slice)
            else:
                # return full chunk
                if not self.chunk_per_page:
                    yield self._get_chunk()

        # return remaining chunk
        if not self.chunk_per_page:
            yield self._get_chunk()

    @require_extension
    def _get_chunk(self):

        if self.offset >= self.stop_dt:
            # terminate sequence
            return None
        # get chunk data based on time
        chunk_delta = self.channel._page_delta(self.chunk_size)
        end = self.offset + datetime.timedelta(microseconds=chunk_delta-1)
        chunk_data = self.chunk.loc[:end]
        # leave remainder
        start = end + datetime.timedelta(microseconds=1)
        self.chunk = self.chunk.loc[start:]
        self.offset = start
        if len(chunk_data):
            # valid data
            return chunk_data
        else:
            # empty chunk
            return pd.Series([])

    @as_native_str()
    def __repr__(self):
        return "<ChannelIterator channel='{}' range=({},{})>".format(
                    self.channel.id, self.start, self.stop)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Time Series API
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class TimeSeriesAPI(APIBase):
    base_uri = "/timeseries"
    name = 'timeseries'

    # ~~~~~~~~~~~~~~~~~~~
    # Channels
    # ~~~~~~~~~~~~~~~~~~~

    def create_channel(self, ts, channel):
        """
        Adds a channel to a timeseries package on the platform.
        """
        if channel.exists:
            return channel
        ts_id = self._get_id(ts)
        resp = self._post( self._uri('/{id}/channels', id=ts_id), json=channel.as_dict())

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = ts.id
        return ch

    def get_channels(self, ts):
        """
        Returns a set of channels for a timeseries package.
        """
        ts_id = self._get_id(ts)
        resp = self._get( self._uri('/{id}/channels', id=ts_id) )

        chs = [TimeSeriesChannel.from_dict(r, api=self.session) for r in resp]
        for ch in chs:
            ch._pkg = ts_id
        return chs

    def get_channel(self, pkg, channel):
        """
        Returns a channel object from the platform.
        """
        pkg_id = self._get_id(pkg)
        channel_id = self._get_id(channel)

        path = self._uri('/{pkg_id}/channels/{id}', pkg_id=pkg_id, id=channel_id)
        resp = self._get(path)

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = pkg_id
        return ch

    def update_channel(self, channel):
        """
        Updates a channel on the platform.

        Note: must be super-admin.
        """
        ch_id = self._get_id(channel)
        pkg_id = self._get_id(channel._pkg)
        path = self._uri('/{pkg_id}/channels/{id}', pkg_id=pkg_id, id=ch_id)

        resp = self._put(path , json=channel.as_dict())

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = pkg_id
        return ch

    def update_channel_properties(self, channel):
        """
        Updates a channel's properties on the platform.

        Note: must be super-admin.
        """
        ch_id = self._get_id(channel)
        pkg_id = self._get_id(channel._pkg)
        path = self._uri('/{pkg_id}/channels/{id}/properties', pkg_id=pkg_id, id=ch_id)

        resp = self._put(path , json=[m.as_dict() for m in channel.properties])

        ch = TimeSeriesChannel.from_dict(resp, api=self.session)
        ch._pkg = pkg_id
        return ch

    def delete_channel(self, channel):
        """
        Deletes a timeseries channel on the platform.
        """

        ch_id = self._get_id(channel)
        pkg_id = self._get_id(channel._pkg)
        path = self._uri('/{pkg_id}/channels/{id}', pkg_id=pkg_id, id=ch_id)

        return self._del(path)

    def delete_channel_by_id(self, pkg_id, channel_id):
        """
        Deletes a timeseries channel on the platform.
        """

        path = self._uri('/{pkg_id}/channels/{id}', pkg_id=pkg_id, id=channel_id)
        return self._del(path)

    # ~~~~~~~~~~~~~~~~~~~
    # Data
    # ~~~~~~~~~~~~~~~~~~~

    @require_extension
    def get_ts_data_iter(self, ts, start, end, channels, chunk_size,
                         use_cache, length=None):
        """
        Iterator will be constructed based over timespan (start,end) or (start, start+seconds)

        Both :chunk_size and :length should be described using strings, e.g.
          5 second  = '5s'
          3 minutes = '3m'
          1 hour    = '1h'
        otherwise microseconds assumed.
        """
        if isinstance(ts, string_types):
            # assumed to be package ID
            ts = self.session.core.get(ts)

        #CHANNELS
        ts_channels = ts.channels

        #no channels specified
        if channels is None:
            channels = ts.channels
        #1 channel specified as TSC object
        elif isinstance(channels,TimeSeriesChannel):
            channels = [channels]
        #1 channel specified and channel id
        elif isinstance(channels,string_types):
            channels = [ch for ch in ts.channels if ch.id==channels]
        #list of channel ids OR ts channels
        else:
            all_ch = []
            for chan in channels:
                if isinstance(chan,string_types):
                    all_ch.extend([ch for ch in ts_channels if ch.id==chan])
                else:
                    all_ch.extend([ch for ch in ts_channels if ch==chan])
            channels = all_ch

        # determine start (usecs)
        the_start = ts.start if start is None else infer_epoch(start)

        # chunk
        if chunk_size is not None and isinstance(chunk_size, string_types):
            chunk_size = parse_timedelta(chunk_size)

        # determine end
        if length is not None:
            if isinstance(length, string_types):
                length_usec = parse_timedelta(length)
            else:
                length_usec = length
            the_end = the_start + length_usec

        elif end is not None:
            the_end = infer_epoch(end)
        else:
            the_end = ts.end

        # logical check
        if the_end < the_start:
            raise Exception("End time cannot be before start time.")

        # loop through chunks
        the_start = int(the_start)
        the_end = int(the_end)

        channel_chunks = [
            ChannelIterator(ch, the_start, the_end, chunk_size,
                            api=self.session, use_cache=use_cache).get_chunks()
            for ch in channels
        ]

        while True:
            # get chunk for all channels
            values = [next(i, None) for i in channel_chunks]
            # no more results?
            if not [1 for v in values if v is not None]:
                break
            # make dataframe
            data_map = {c.name: v for c,v in zip(channels,values) if v is not None}
            yield pd.DataFrame.from_dict(data_map)

    @require_extension
    def get_ts_data(self, ts, start, end, length, channels, use_cache):
        """
        Retrieve data. Must specify end-time or length.
        """
        ts_iter = self.get_ts_data_iter(ts=ts, start=start, end=end, channels=channels,
                                         chunk_size=None, use_cache=use_cache, length=length)
        df = pd.DataFrame()
        for tmp_df in ts_iter:
            df = df.append(tmp_df)
        return df

    def get_segments(self, ts, channel, start, stop, gap_factor):
        """
        Retrieve ranges of time for channel (between start and stop)
        where there exists contiguous data. Gap detection sensitivity
        can be adjusted using ``gap_factor`` which is multiplied
        with sampling period of channel for identifying gaps.
        """
        channel_id = self._get_id(channel)
        package_id = self._get_id(ts)
        start = infer_epoch(start)
        stop  = infer_epoch(stop)

        resp = self._get(
            # Note: uses streaming server
            host     = self.session._host,
            endpoint = '/streaming/ts/retrieve/segments',
            base     = '',
            stream   = True,
            params   = dict(
                channel = channel_id,
                package = package_id,
                session = self.session.token,
                start   = start,
                end     = stop,
                gapThreshold = gap_factor
            )
        )
        return [tuple(x) for x in resp]

    # ~~~~~~~~~~~~~~~~~~~
    # Annotation Layers
    # ~~~~~~~~~~~~~~~~~~~

    def create_annotation_layer(self, ts, layer, description):

        if isinstance(layer,TimeSeriesAnnotationLayer):
            data = layer.as_dict()
        elif isinstance(layer, string_types):
            data = {
                'name' : layer,
                'description' : description
            }
        else:
            raise Exception("Layer must be TimeSeriesAnnotationLayer object or name of new layer")

        existing_layer = [i for i in ts.layers if data['name'] == i.name]
        if existing_layer:
            print('Returning existing layer {}'.format(existing_layer))
            return existing_layer[0]
        else:
            ts_id = self._get_id(ts)
            path = self._uri('/{id}/layers', id=ts_id)
            resp = self._post(path, json=data)
            tmp_layer = TimeSeriesAnnotationLayer.from_dict(resp, api=self.session)
            if isinstance(layer,TimeSeriesAnnotationLayer):
                layer.__dict__.update(tmp_layer.__dict__)
            return tmp_layer

    def get_annotation_layer(self, ts, layer):
        ts_id = self._get_id(ts)
        layer_id = self._get_id(layer)
        path = self._uri('/{id}/layers/{layer_id}',id=ts_id,layer_id=str(layer_id))
        resp = self._get(path)
        return TimeSeriesAnnotationLayer.from_dict(resp, api=self.session)

    def get_annotation_layers(self, ts):
        ts_id = self._get_id(ts)
        resp = self._get(self._uri('/{id}/layers', id=ts_id))
        return [TimeSeriesAnnotationLayer.from_dict(x, api=self.session) for x in resp["results"]]

    def update_annotation_layer(self, ts, layer):
        #return all layers
        ts_id = self._get_id(ts)
        layer_id = self._get_id(layer)
        path = self._uri('/{id}/layers/{layer_id}', id=ts_id, layer_id=layer_id)
        resp = self._put(path, json=layer.as_dict())
        return TimeSeriesAnnotationLayer.from_dict(resp, api=self.session)

    def delete_annotation_layer(self, layer):
        ts_id = layer.time_series_id
        path = self._uri('/{id}/layers/{layer_id}',id = ts_id, layer_id =layer.id)
        try:
            self._del(path)
            layer.id = None
            return True
        except requests.exceptions.HTTPError:
            return False

    # ~~~~~~~~~~~~~~~~~~~
    # Annotations
    # ~~~~~~~~~~~~~~~~~~~

    def delete_annotation(self, annot):
        """
        Deletes a single annotation
        """
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations/{annot_id}',
                    ts_id = annot.time_series_id,
                    layer_id = annot.layer_id,
                    annot_id = annot.id)
        try:
            self._del(path)
            annot.id = None
            return True
        except requests.exceptions.HTTPError:
            return False

    def create_annotations(self,layer, annotations):

        all_annotations = []

        if not isinstance(annotations,list):
            annotations = [annotations]

        for annot in annotations:
            tmp = self.create_annotation(layer=layer,annotation=annot)
            all_annotations.append(tmp)

        #if adding single annotation, return annotation object, else return list
        if len(all_annotations) == 1:
            all_annotations = all_annotations[0]

        return all_annotations

    def create_annotation(self, layer, annotation, **kwargs):
        """
        Creates annotation for some timeseries package on the platform.
        """
        if isinstance(annotation,TimeSeriesAnnotation):
            data = annotation.as_dict()
        elif all(x in kwargs for x in ['start','end']):
            start_time = infer_epoch(kwargs['start'])
            end_time = infer_epoch(kwargs['end'])
            data = {
                'name':'',
                'label':annotation,
                'start':int(start_time),
                'end':int(end_time),
            }
            if kwargs['channel_ids']:
                channel_ids = kwargs['channel_ids']
                if isinstance(channel_ids,string_types):
                    channel_ids = [channel_ids]
                data['channelIds']=channel_ids
            else:
                ts = layer._api.core.get(layer.time_series_id)
                data['channelIds']=[x.id for x in ts.channels]
            if 'description' in annotation:
                data['description']=kwargs['description']
            else:
                data['description']=None
        else:
            raise Exception("Must provide TimeSeriesAnnotation object or 'annotation','start','end' at minimum")

        data['time_series_id'] = layer.time_series_id
        data['layer_id'] = layer.id

        path = self._uri('/{ts_id}/layers/{layer_id}/annotations',
                    ts_id=layer.time_series_id, layer_id=layer.id)
        resp = self._post(path, json=data)
        tmp = TimeSeriesAnnotation.from_dict(resp, api=self.session)

        if isinstance(annotation,TimeSeriesAnnotation):
            annotation.__dict__.update(tmp.__dict__)

        return tmp

    def update_annotation(self, ts, layer, annot):
        """
        Update annotation on the platform.
        """
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations/{annot_id}',
                    ts_id = self._get_id(ts),
                    layer_id = self._get_id(layer),
                    annot_id = self._get_id(annot))
        resp = self._put(path, json=annot.as_dict())
        return TimeSeriesAnnotation.from_dict(resp, api=self.session)

    def get_annotation(self, ts, layer, annot):
        """
        Returns a timeseries annotation
        """
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations/{annot_id}',
                    ts_id = self._get_id(ts),
                    layer_id = self._get_id(layer),
                    annot_id = self._get_id(annot))
        resp = self._get(path)
        return TimeSeriesAnnotation.from_dict(resp["annotation"], api=self.session)

    def iter_annotations(self, ts, layer, window_size=10, channels=None):
        # window_size is seconds

        if not isinstance(ts, TimeSeries):
            raise Exception("Argument 'ts' must be TimeSeries.")

        # paginate annotations
        start_time, end_time = ts.limits()
        num_windows = (end_time-start_time)/(window_size*1e6)
        for i in range(int(math.ceil(num_windows))):
            win_start = start_time + i* (window_size*1e6)
            win_end = win_start + window_size*1e6
            if win_end > end_time:
                win_end = end_time
            yield self.get_annotations(ts=ts, layer=layer, start=win_start, end=win_end, channels=channels)

    def get_annotations(self, ts, layer, start=None, end=None, channels=None):
        """
        Returns all annotations for a given layer
        """
        limit = 100
        annots = []
        for offset in itertools.count(0, limit):
            batch = self.query_annotations(
                ts=ts, layer=layer, channels=channels, start=start, end=end,
                limit=limit, offset=offset)
            if not batch:
                break
            annots += batch

        return annots

    def requested_channels(self, ts, channels):
        # empty uses all channels
        if channels is None:
            return []

        ch_list = [self._get_id(x) for x in channels]

        # validate
        all_channels = set([ch.id for ch in ts.channels])
        if not set(ch_list).issubset(all_channels):
            raise Exception("Channels {chs} not found in TimeSeries '{ts}'".format(
                  ts=ts.id, chs=list(all_channels - set(ch_list))))

        return ch_list

    def query_annotations(self, ts, layer, start=None, end=None, channels=None, limit=100, offset=0):
        """
        Retrieves timeseries annotations for a particular range  on array of channels.
        """
        ch_list = self.requested_channels(ts, channels)

        ts_start, ts_end = ts.limits()
        if start is None:
            start = ts_start
        elif isinstance(start, datetime.datetime):
            start = usecs_since_epoch(start)

        if end is None:
            end = ts_end
        elif isinstance(end, datetime.datetime):
            end = usecs_since_epoch(end)

        params = {
            'start': int(start),
            'end': int(end),
            'channelIds': ch_list,
            'layerName': layer.name,
            'limit': limit,
            'offset': offset
        }
        path = self._uri('/{ts_id}/layers/{layer_id}/annotations',
                    ts_id = ts.id,layer_id = layer.id)

        resp = self._get(path, params=params)

        return [TimeSeriesAnnotation.from_dict(x, api=self.session) for x in resp['annotations']['results']]

    def query_annotation_counts(self, ts, layers, start, end, period, channels=None, merge_periods=False):
        """
        Retrieves annotation counts for a given ts, channel, start, end, and/or layer

        Args:
            ts (TimeSeries)                  : The timeseries package for which to count annotations
            layers ([TimeSeriesLayer])       : List of layers for which to count annotations
            start (datetime or microseconds) : The starting time of the range to query
            end (datetime or microseconds)   : The ending time of the the range to query
            period (string)                  : The length of time to group the counts.
                                               Formatted as a string - e.g. '1s', '5m', '3h'
            channels ([TimeSeriesChannel])   : List of channel (if omitted, all channels will be used)
            merge_periods(Boolean)           : If true, merge consecutive result periods together to
                                               reduce the size of the resulting payload

        Returns:
            A dict
            layer_id -> list of counts for each period
        """
        ch_list = self.requested_channels(ts, channels)

        if isinstance(start, datetime.datetime):
            start = usecs_since_epoch(start)
        if isinstance(end, datetime.datetime):
            end = usecs_since_epoch(end)

        period = parse_timedelta(period)

        params = {
            'aggregation': 'count',
            'start': int(start),
            'end': int(end),
            'period': int(period),
            'mergePeriods': merge_periods,
            'layerIds': [l.id for l in layers],
            'channelIds': ch_list
        }

        path = self._uri('/{ts_id}/annotations/window',
                    ts_id = self._get_id(ts))

        resp = self._get(path, params=params)
        return resp

    @require_extension
    def process_annotation_file(self,ts,file_path):
        """
        Processes the .bfannot file at file_path and adds to timeseries package
        """
        if not file_path.lower().endswith(('.bfannot')):
            raise Exception("Annotation file format not currently supported. Supported annotations types: .bfannot")
        try:
            df = pd.read_csv(file_path)
            df = df.where((pd.notnull(df)), None)
            channels = ts.channels
            if (df['version'][0] == 1.0): #version number
                layers = df['layer_name'].unique()
                for l in layers:
                    layer = []
                    annots = df.loc[df['layer_name'] == l]

                    #create or find existing layer
                    layer = ts.add_layer(layer=l, description=annots['layer_description'].iloc[0])

                    annotations = []
                    for index, row in annots.iterrows():
                        if pd.isnull(row['channel_names']):
                            channel_ids = [x.id for x in channels]
                        else:
                            channel_names = row['channel_names'].split(';')
                            channel_ids = [x.id for x in channels if x.name in channel_names]

                        layer.insert_annotation(annotation=row['annotation_label'],channel_ids=channel_ids,start=row['start_uutc'],end=row['end_uutc'],description=row['annotation_description'])

                    print('Added annotations to layer {} , pkg: {}'.format(layer,ts))
            else:
                raise Exception('BF version {} not found or not supported'.format(df['version'][0]))

        except Exception as error:
            raise Exception("Error adding annotation file {}, {}".format(file_path, error))

    @require_extension
    def write_annotation_file(self,ts,file_path,layer_names):
        """
        Writes all layers in ts to .bfannot (v1.0) file

        """
        layers = ts.layers
        if not layers:
            raise Exception('Timeseries has no existing layers')

        if not file_path.lower().endswith(('.bfannot')):
            file_path+='.bfannot'

        headers = ['version',
                    'package_type',
                    'layer_name',
                    'layer_description',
                    'annotation_label',
                    'start_uutc',
                    'end_uutc',
                    'channel_names',
                    'annotation_description']
        out = pd.DataFrame(columns=headers)

        if layer_names:
            if not isinstance(layer_names,list):
                layer_names = [layer_names]
            new_layers = [l for l in layers if l.name in layer_names]
            layers = new_layers

        to_write = []
        for l in layers:
            annot = l.annotations()
            for a in annot:
                channels = [ch.name for ch in ts.channels if ch.id in a.channel_ids]
                channel_names = ";".join(channels)
                tmp = {
                    'layer_name' : l.name,
                    'layer_description': l.description,
                    'annotation_label' : a.label,
                    'start_uutc' : int(a.start),
                    'end_uutc' : int(a.end),
                    'channel_names' : channel_names,
                    'annotation_description' : a.description
                }
                to_write.append(tmp)
        out = out.append(to_write)

        # Add version number and package type
        out.loc[0,('version')] = 1.0
        out.loc[0,('package_type')] = ts.type

        out.to_csv(file_path,index=False)

    # ~~~~~~~~~~~~~~~~~~~
    # Helpers
    # ~~~~~~~~~~~~~~~~~~~


    def _annotation_query_params(ts, start, end, period, layer, channels):
        # parse channel input
        channels = self._channel_list(ts, channels)

        # channel IDs
        ch_ids = [x.id for x in channels]

        params = {
            'start': usecs_since_epoch(start),
            'end': usecs_since_epoch(end),
            'channels': ch_ids,
            'period': period
        }

        if layer is not None:
            params['layer'] = layer
        return params

    def _channel_list(self, ts, channels):
        """
        Get list of channel objects provided flexible input values
        """
        ts_id = self._get_id(ts)

        if channels is None:
            # if channel(s) not specified, grab all for package
            channels = self.session.timeseries.get_channels(ts_id)

        # check if list
        if not hasattr(channels, '__iter__'):
            # they specified a single object
            channels = [channels]

        # check type of items in list
        for ch in channels:
            if isinstance(ch, TimeSeriesChannel):
                # Channel looks good
                continue
            if isinstance(ch, string_types):
                # Assume channel ID, get object
                ch = self.session.get(ch)
            else:
                raise Exception('Expecting TimeSeries instance or ID')

        return channels



import functools


try:
    import pandas
except ImportError:
    pandas = None

try:
    import numpy
except ImportError:
    numpy = None


class MissingDependency(Exception):
    pass


def require_extension(f):
    """
    Decorator that ensures the optional `data` dependencies are installed.

    All functions and methods that use `numpy` or `pandas` must be decorated
    with this.
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if pandas is None or numpy is None:
            raise MissingDependency("""This command require additional dependencies. To install, run:

pip install blackfynn[data]

""")
        return f(*args, **kwargs)
    return wrapper



from __future__ import division

import errno
import json
import os
import socket
import subprocess
import sys
from collections import OrderedDict
from time import sleep

import semver
from future.utils import raise_from
from blackfynn.log import get_logger
from blackfynn.models import Collection, Dataset, DataPackage

logger = get_logger('blackfynn.agent')

try:
    from websocket import create_connection
except ModuleNotFoundError:
    logger.warn("websocket-client is not installed - uploading with the Agent will not work")


MINIMUM_AGENT_VERSION = semver.parse_version_info("0.2.2")
DEFAULT_LISTEN_PORT = 11235


class AgentError(Exception):
    pass


def agent_cmd():
    if sys.platform == 'darwin':
        return '/usr/local/opt/blackfynn/bin/blackfynn_agent'

    elif sys.platform.startswith('linux'):
        return '/opt/blackfynn/bin/blackfynn_agent'

    elif sys.platform in ['win32', 'cygwin']:
        return 'C:/Program Files/Blackfynn/blackfynn_agent.exe'

    raise AgentError('Platform {} is not supported'.format(sys.platform))


def validate_agent_installation(settings):
    """
    Check whether the agent is installed and at least the minimum version.
    """
    try:
        version = subprocess.check_output([agent_cmd(), 'version'], env=agent_env(settings))
    except (AgentError, subprocess.CalledProcessError, EnvironmentError) as e:
        raise AgentError('Agent not installed. Visit https://developer.blackfynn.io/agent for installation directions.')

    try:
        agent_version = semver.parse_version_info(version.decode().strip())
    except ValueError as e:
        raise_from(AgentError('Invalid version string'), e)

    if agent_version < MINIMUM_AGENT_VERSION:
        raise AgentError('Agent not compatible: found version {}, need version {}'.format(
                         agent_version, MINIMUM_AGENT_VERSION))

    logger.info('Agent version %s found', agent_version)


def agent_env(settings):
    """
    Configure the agent environment to mirror the Python client
    The "local" environment looks for the host in BLACKFYNN_API_LOC
    (this is configured down in blackfynn-rust)
    """
    env = {
        'BLACKFYNN_API_ENVIRONMENT': 'local',
        'BLACKFYNN_API_LOC': settings.api_host,
        'BLACKFYNN_API_TOKEN': settings.api_token,
        'BLACKFYNN_API_SECRET': settings.api_secret,
    }
    if sys.platform in ['win32', 'cygwin']:
        env['SYSTEMROOT'] = os.getenv('SYSTEMROOT')
     # On Windows, the SYSTEMROOT environment variable must be preserved for DLLs to correctly load.
     # ref: https://travis-ci.community/t/socket-the-requested-service-provider-could-not-be-loaded-or-initialized/1127

    if 'BLACKFYNN_LOG_LEVEL' in os.environ:
        env['BLACKFYNN_LOG_LEVEL'] = os.environ.get('BLACKFYNN_LOG_LEVEL')

    logger.debug("Agent environment: %s", env)

    return env


class AgentListener(object):
    """
    Context manager that starts the agent in listen server mode.
    """
    def __init__(self, settings, port):
        self.settings = settings
        self.port = port
        self.proc = None
        self.devnull = None

    def __enter__(self):
        check_port(self.port)
        command = [agent_cmd(), 'upload-status', '--listen', '--port', str(self.port)]

        self.devnull = open(os.devnull, 'w')
        self.proc = subprocess.Popen(command, env=agent_env(self.settings),
                                     stdout=self.devnull, stderr=self.devnull)
        return self.proc

    def __exit__(self, *exc):
        self.proc.kill()
        self.devnull.close()


def check_port(port):
    """
    Refuse to start up if the agent is already running in listen mode.
    This can cause problems with relative files paths and session credentials.
    """
    try:
        logger.debug("Checking port %s", port)
        create_connection(socket_address(port)).close()
    except socket.error as e:
        if e.errno == errno.ECONNREFUSED:  # ConnectionRefusedError for Python 3
            logger.debug("No agent found, port %s OK", port)
            return True
        else:
            raise
    else:
        raise AgentError("The agent is already running. Please stop any running processes and try again")


def socket_address(port):
    return "ws://localhost:{}".format(port)


def create_agent_socket(port):
    """
    Open a websocket connection to the agent

    If the agent is not available, wait using exponential backoff for it to
    come up and start responding to messages
    """
    for i in range(-2, 4):
        try:
            return create_connection(socket_address(port))
        except socket.error as e:
            if e.errno == errno.ECONNREFUSED:  # ConnectionRefusedError for Python 3
                sleep_time = 2 ** i
                logger.debug("Connection refused - sleeping for %s seconds", sleep_time)
                sleep(sleep_time)
            else:
                raise

    raise AgentError("Could not connect to Agent")


def agent_upload(destination, files, dataset, append, recursive, display_progress, settings):
    """
    Push an upload through the agent.
    """
    directory_upload = any(os.path.isdir(f) for f in files)

    if directory_upload and len(files) > 1:
        raise AgentError(
            'Can only upload a single directory.\n'
            'Please pass a single directory argument: `pkg.upload("/experiment/dir")`')

    if recursive and not directory_upload:
        raise AgentError(
            'Recursive uploads are only allowed with directories.\n'
            'Upload a directory or pass `recursive=False`.')

    if recursive and append:
        raise AgentError('Cannot use `recursive=True` when appending`')

    # Figure out what files the agent is going to upload.
    # We cannot count on the agent to send "upload queued" messages for
    # all files before it starts uploading, so we generate the files we
    # plan to wait for.
    if directory_upload:
        directory = files[0]
        if recursive:
            expected_files = []
            for dirpath, _, filenames in os.walk(directory):
                for f in filenames:
                    expected_files.append(os.path.join(dirpath, f))
        else:
            expected_files = []
            for f in os.listdir(directory):
                path = os.path.join(directory, f)
                if os.path.isfile(path):
                    expected_files.append(path)
    else:
        expected_files = files

    # Agent uses absolute paths
    expected_files = [os.path.abspath(f) for f in expected_files]

    if isinstance(destination, Dataset):
        dataset_id = destination.id
        package_id = None
    elif isinstance(destination, (Collection, DataPackage)):
        dataset_id = dataset.id
        package_id = destination.id
    else:
        raise ValueError('Can only upload to a Dataset, Package, or Collection')

    with AgentListener(settings, DEFAULT_LISTEN_PORT):
        try:
            ws = create_agent_socket(DEFAULT_LISTEN_PORT)

            ws.send(json.dumps({
                'message': 'queue_upload',
                'body': {
                    'dataset': dataset_id,
                    'package': package_id,
                    'files': files,
                    'append': append,
                    'recursive': recursive
            }}))

            upload_manager = UploadManager(expected_files, display_progress)
            upload_manager.print_progress()

            for msg in ws:
                msg = json.loads(msg)

                if msg['message'] == 'file_queued_for_upload':
                    upload_manager.set_queued(msg['path'], msg['import_id'])

                elif msg['message'] == 'upload_progress':
                    upload_manager.set_progress(msg['path'], msg['import_id'],
                                                msg['percent_done'], msg['done'])

                elif msg['message'] == 'upload_complete':
                    upload_manager.set_complete(msg['import_id'])

                elif msg['message'] == 'upload_error':
                    logger.error(msg['context'])
                    upload_manager.set_error(msg['import_id'])

                elif msg['message'] == 'error':
                    raise AgentError(msg['context'])

                else:
                    logger.debug("Unknown message", msg)

                upload_manager.print_progress()
                if upload_manager.done:
                    break

        finally:
            try:
                ws.close()
            except UnboundLocalError:
                pass

def remove_prefix(text, prefix):
    return text[text.startswith(prefix) and len(prefix):]

class UploadManager(object):
    """
    Manager for file status and messages.

    This is complicated by that fact that the agent sends status information
    for files that are already in the queue or started by other processes.
    This makes it possible for the same file to be queued twice, so we
    have to track both the filename and import id. We only want to wait for
    all "our" files to upload.
    """
    def __init__(self, files, display_progress):
        # Should we show progress bars?
        self.display_progress = display_progress

        # map of filepath -> list(FileProgress)
        self.uploads = OrderedDict()

        for file in files:
            self.track_file(file, import_id=None, ours=True)

        # Keep track of whether progress bars have already been rendered so
        # we know if/what to erase when re-drawing
        self.lines_on_screen = 0

    def track_file(self, file, import_id, ours):
        progress = FileProgress(file, import_id, ours)
        if file in self.uploads:
            self.uploads[file].append(progress)
        else:
            self.uploads[file] = [progress]
        return progress

    def get_tracked_file(self, file, import_id):
        if sys.platform in ['win32', 'cygwin']:
            file = remove_prefix(file, "\\\\?\\") #windows OS sometimes prefix the filepath with \\?\, so we remove it if we see if
        if file in self.uploads:
            for p in self.uploads[file]:
                if p.import_id == import_id:
                    return p

    def all_tracked_files(self):
        for filegroup in self.uploads.values():
            for progress in filegroup:
                yield progress

    def set_queued(self, file, import_id):
        # Update the unqueued version of the file with an import id
        progress = self.get_tracked_file(file, None)

        # This import is queued from a different process - we don't care
        if progress is None:
            return

        progress.queued = True
        progress.import_id = import_id

    def set_progress(self, file, import_id, percent_done, done):
        # Absorb any files that are in the DB/already queued
        if self.get_tracked_file(file, import_id) is None:
            self.track_file(file, import_id, ours=False)

        progress = self.get_tracked_file(file, import_id)
        progress.percent_done = percent_done

    def set_complete(self, import_id):
        for progress in self.all_tracked_files():
            if progress.import_id == import_id:
                progress.done = True

    def set_error(self, import_id):
        for progress in self.all_tracked_files():
            if progress.import_id == import_id:
                progress.errored = True

    @property
    def done(self):
        return all([fstat.done for fstat in self.all_tracked_files() if fstat.ours])

    def print_progress(self, width=24):
        if not self.display_progress:
            return

        # move cursor to relative beginning
        sys.stdout.write("\033[F" * self.lines_on_screen)

        for fstat in self.all_tracked_files():
            if fstat.done:
                state = 'DONE'
            elif fstat.errored:
                state = 'ERRORED'
            elif fstat.queued:
                state = 'UPLOADING'
            else:
                state = 'WAITING'

            text = ' [ {bars}{dashes} ] {state:12s} {percent:05.1f}% {name}\n'.format(
                bars='#' * int(fstat.progress * width),
                dashes='-' * (width - int(fstat.progress * width)),
                percent=fstat.percent_done,
                name=fstat.name,
                state=state)

            sys.stdout.write('{}\r'.format(text))
            sys.stdout.flush()

        self.lines_on_screen = len(list(self.all_tracked_files()))


class FileProgress(object):
    def __init__(self, filename, import_id, ours):
        # We only care about the state of uploads started by this process
        self.ours = ours
        self.filename = filename
        self.import_id = import_id
        self.name = os.path.basename(filename)
        self._percent_done = 0
        self.done = False
        self.errored = False
        self.queued = False

    @property
    def percent_done(self):
        if self.done:
            return 100
        return self._percent_done

    @percent_done.setter
    def percent_done(self, value):
        # Only increment progress (in case messages come out of order)
        if value > self._percent_done:
            self._percent_done = value

    @property
    def progress(self):
        return self.percent_done / 100



import uuid
import pytest

from pkg_resources import resource_filename

from blackfynn.models import TimeSeries
from blackfynn.api.agent import AgentError


def _resource_path(fname):
    return resource_filename('tests.resources', fname)

# All test assets need to use this "test-78f3ea50" prefix so failures caused by
# deleting datasets before processing all assests can be filtered out of
# production error notifications.
FILE1      = _resource_path('test-78f3ea50.txt')
FILE2      = _resource_path('test-78f3ea50.csv')
FILE3      = _resource_path('test-78f3ea50.png')
FILE_EMPTY = _resource_path('test-78f3ea50.empty')
FLAT_DIR   = _resource_path('flat_dir')
NESTED_DIR = _resource_path('nested_dir')
INNER_DIR  = 'inner_dir'


def test_upload_legacy_to_dataset(dataset):
    n = len(dataset.items)
    dataset.upload(FILE1, display_progress=True, use_agent=False)
    dataset.update()

    # Check that we've added an item to the dataset root
    assert n + 1 == len(dataset.items)


@pytest.mark.parametrize('upload_args,n_files', [
    ([FILE1], 1),               # Single file
    ([[FILE1, FILE2]], 2),      # Multiple files
])
def test_upload_legacy_to_collection(dataset, upload_args, n_files):
    collection = dataset.create_collection(str(uuid.uuid4()))
    collection.upload(*upload_args, use_agent=False)

    assert len(collection.items) == n_files


@pytest.mark.agent
@pytest.mark.parametrize('upload_args, n_files', [
    ([FLAT_DIR], 3),
    ([NESTED_DIR], 1),
    ([NESTED_DIR + '/' + INNER_DIR], 2),
])
def test_upload_directory(dataset, upload_args, n_files):
    collection = dataset.create_collection(str(uuid.uuid4()))
    collection.upload(upload_args)
    assert len(collection.items) == n_files


@pytest.mark.agent
def test_upload_recursive(dataset):
    collection = dataset.create_collection(str(uuid.uuid4()))
    collection.upload(NESTED_DIR, recursive=True)

    assert len(collection.items) == 1
    nested_dir = collection.items[0]
    assert len(nested_dir.items) == 2
    inner_dir = next(pkg for pkg in nested_dir.items if pkg.name == INNER_DIR)
    assert len(inner_dir.items) == 2


@pytest.mark.agent
def test_upload_recursive_flag_is_not_allowed_with_file(dataset):
    collection = dataset.create_collection(str(uuid.uuid4()))
    with pytest.raises(AgentError):
        collection.upload(FILE1, recursive=True)


@pytest.mark.agent
def test_upload_cannot_upload_multiple_directories(dataset):
    with pytest.raises(AgentError):
        dataset.upload(FLAT_DIR, FILE1)


@pytest.mark.agent
@pytest.mark.parametrize('upload_args,n_files', [
    ([FILE1], 1),          # Single file
    ([FILE1, FILE2], 2),   # Multiple files, separate arguments
    ([[FILE1, FILE2]], 2), # Multiple files, single argument list
])
def test_upload_to_dataset(dataset, upload_args, n_files):
    """
    Note: ETL will fail since destination will likely be removed
          before being processed.
    """
    c = len(dataset.items)
    dataset.upload(*upload_args)
    dataset.update()
    assert len(dataset.items) == c + n_files


@pytest.mark.parametrize('append_args,n_files', [
    ([FILE1], 1),          # Single file
    ([FILE1, FILE2], 2),   # Multiple files, separate arguments
    ([[FILE1, FILE2]], 2), # Multiple files, single argument list
])
def test_append_legacy(dataset, append_args, n_files):
    # TimeSeries package to append into...
    pkg = TimeSeries('Rando Timeseries')
    dataset.add(pkg)

    resp = pkg.append_files(*append_args, use_agent=False)
    assert len(resp) == n_files

    for r in resp:
        manifest = r[0]['manifest']
        assert manifest['content'] is not None
        assert manifest['type'] == 'append'


@pytest.mark.parametrize('upload_args,n_files', [
    ([FILE1], 1),          # Single file
    ([FILE1, FILE2], 2),   # Multiple files, separate arguments
    ([[FILE1, FILE2]], 2), # Multiple files, single argument list
])
def test_upload_legacy(dataset, upload_args, n_files):
    """
    Note: ETL will fail since destination will likely be removed
          before being processed.
    """
    resp = dataset.upload(*upload_args, use_agent=False)
    assert len(resp) == n_files

    for r in resp:
        manifest = r[0]['manifest']
        assert manifest['content'] is not None
        assert manifest['type'] == 'upload'


def test_cannot_upload_directory_using_s3(dataset):
    with pytest.raises(Exception):
        dataset.upload(FLAT_DIR, use_agent=False)


@pytest.mark.agent
@pytest.mark.parametrize('append_args,n_files', [
    ([FILE1], 1),          # Single file
    ([FILE1, FILE2], 2),   # Multiple files, separate arguments
    ([[FILE1, FILE2]], 2), # Multiple files, single argument list
])
def test_append(dataset, append_args, n_files):
    """
    Note: ETL will fail for append, because... it's a text file.
          But also because the destination will likely be removed
          before the ETL actually places the new node there.
    """
    # TimeSeries package to append into...
    pkg = TimeSeries('Rando Timeseries')
    dataset.add(pkg)

    # upload/append file into package
    pkg.append_files(*append_args)
    #TODO: assert append was successful


@pytest.mark.agent
def test_progress_for_empty_files(dataset):
    dataset.upload(FILE_EMPTY, display_progress=True)



import datetime
from collections import namedtuple
from past.builtins import unicode  # Alias of str in Python 3

import pytest

from blackfynn.models import (
    DataPackage,
    ModelProperty,
    ModelPropertyEnumType,
    ModelPropertyType,
)
from tests.utils import create_test_dataset, current_ts, get_test_client


@pytest.mark.parametrize('from_type,blackfynn_type,data_type', [
    (int, 'long', int),
    (ModelPropertyType(data_type=int), 'long', int),
    (ModelPropertyType(data_type=unicode), 'string', unicode),
    (datetime.date, 'date', datetime.date),
    ('date', 'date', datetime.datetime),
    (ModelPropertyType(data_type=str, format='date'), 'string', str),
])
def test_model_type_conversion(from_type, blackfynn_type, data_type):
    property_type = ModelPropertyType._build_from(from_type)
    assert property_type.data_type == data_type
    assert property_type._blackfynn_type == blackfynn_type


@pytest.mark.parametrize('json,data_type,format,unit', [
    ('string', unicode, None, None),
    ({'type':'string', 'format':'date'}, unicode, 'date', None),
    ({'type': 'double', 'unit': 'kg'}, float, None, 'kg'),
])
def test_model_type_from_dict(json, data_type, format, unit):
    decoded = ModelPropertyType._build_from(json)
    assert isinstance(decoded, ModelPropertyType)
    assert decoded.data_type == data_type
    assert decoded.format == format
    assert decoded.unit == unit


def test_model_with_invalid_properties(dataset):
    invalid_schema = [('an_integer', int, 'An Integer')]
    new_model = dataset.create_model(
        'New_Model_{}'.format(current_ts()), 'A New Model', 'a new model', invalid_schema
    )

    assert not new_model.schema


def test_date_formatting():
    d1 = datetime.datetime(2018, 8, 24, 15, 11, 25)
    assert ModelPropertyType(datetime.datetime)._encode_value(d1) == '2018-08-24T15:11:25.000000+00:00'

    d2 = datetime.datetime(2018, 8, 24, 15, 11, 25, 1)
    assert ModelPropertyType(datetime.datetime)._encode_value(d2) == '2018-08-24T15:11:25.000001+00:00'

def test_boolean_string():
    assert ModelPropertyType(bool)._decode_value("false") == False
    assert ModelPropertyType(bool)._decode_value("true") == True
    assert ModelPropertyType(bool)._decode_value("nope") == True

def test_models(dataset):
    schema = [('an_integer', int, 'An Integer', True), ('a_bool', bool), ('a_string', str), ('a_datetime', datetime.datetime)]
    display_name = 'A New Property'
    description = 'a new description'
    values = {'an_integer': 100, 'a_bool': True, 'a_string': 'fnsdlkn#$#42nlfds$3nlds$#@$23fdsnfkls', 'a_datetime': datetime.datetime.now()}

    #################################
    ## Models
    ################################

    models = dataset.models()

    new_model = dataset.create_model('New_Model_{}'.format(current_ts()), 'A New Model', 'a new model', schema)

    assert len(dataset.models()) == len(models) + 1

    assert dataset.get_model(new_model.id) == new_model
    assert dataset.get_model(new_model.type) == new_model

    # Check that local changes get propagated
    new_model.add_property('a_new_property', str, display_name)
    new_model.description = description
    new_model.update()
    new_model = dataset.get_model(new_model.id)
    assert new_model.description == description
    assert new_model.get_property('a_new_property').display_name == display_name

    new_model.add_properties([('a_new_float', float), {'name': 'a_new_int', 'data_type': int}, 'a_new_string'])
    assert new_model.get_property('a_new_float').type == float
    assert new_model.get_property('a_new_int').type == int
    assert new_model.get_property('a_new_string').type == unicode

    nc_one = new_model.create_record(values)
    nc_two = new_model.create_record({'an_integer': 1, 'a_bool': False, 'a_string': '', 'a_datetime': datetime.datetime.now()})
    nc_three = new_model.create_record({'an_integer': 10000, 'a_bool': False, 'a_string': '43132312', 'a_datetime': datetime.datetime.now()})
    nc_four = new_model.create_record({'a_datetime': datetime.datetime.now()})

    nc_delete_one = new_model.create_record({'a_datetime': datetime.datetime.now()})
    nc_delete_two = new_model.create_record({'a_datetime': datetime.datetime.now()})

    with pytest.raises(Exception):
        new_model.create_record()

    new_models_old = new_model.get_all()
    assert new_model.get_all(limit=1) == new_models_old[:1]
    assert new_model.get_all(limit=2, offset=2) == new_models_old[2:4]

    new_model.delete_records(nc_delete_one, nc_delete_two.id)
    new_models = new_model.get_all()

    assert len(new_models) == (len(new_models_old) - 2)

    assert nc_two.model == new_model
    assert nc_two.get('a_string') == ''

    nc_four.set('a_string', 'hello')
    assert nc_four.get('a_string') == new_model.get(nc_four).get('a_string')

    with pytest.raises(Exception):
        nc_four.set('an_integer', datetime.datetime.now())

    assert nc_four.get('an_integer') == None
    nc_four.set('an_integer', 10)
    assert nc_four.get('an_integer') == 10

    nc_delete_three = new_model.create_record({'a_string': 'delete me'})
    assert len(new_model.get_all()) == len(new_models) + 1
    nc_delete_three.delete()
    assert len(new_model.get_all()) == len(new_models)

    # cannot add a record id column using an existing name
    with pytest.raises(ValueError):
        new_model.get_all().as_dataframe(
            record_id_column_name=list(new_model.get_all().type.schema.keys())[0]
        )

    # assert no extra columns are added by default
    df_cs_no_rec = new_model.get_all().as_dataframe()
    assert len(df_cs_no_rec.columns) == len(new_model.get_all().type.schema.keys())

    # assert record id column is added when arg is present and valid
    df_cs = new_model.get_all().as_dataframe(
        record_id_column_name='record_id'
    )

    # confirm that all record ids are present in this dataframe
    assert 'record_id' in df_cs.columns
    for record in new_model.get_all():
        assert not df_cs.query('record_id == @record.id').empty

    #################################
    ## Relationships
    ################################

    relationships = dataset.relationships()

    new_relationship = dataset.create_relationship_type('New_Relationship_{}'.format(
        current_ts()), 'a new relationship')

    assert len(dataset.relationships()) == len(relationships) + 1

    assert dataset.get_relationship(new_relationship.id) == new_relationship
    assert dataset.get_relationship(new_relationship.type) == new_relationship

    nr_one = new_relationship.relate(nc_one, nc_two)
    nr_four = new_relationship.relate(nc_four, nc_one)
    nr_five = new_relationship.relate(nc_four, nc_two)

    nr_two = nc_two.relate_to(nc_three, new_relationship)
    nr_three = nc_three.relate_to(nc_four, new_relationship)
    nr_six = nc_four.relate_to(nc_three, new_relationship)
    nr_seven = nc_four.relate_to(nc_one, relationship_type="goes_to")

    assert nr_seven[0].destination == nc_one.id
    assert nr_seven[0].source == nc_four.id
    assert nr_seven[0].type == "goes_to"
    assert len(nc_four.get_related(new_model.type)) == 5

    new_relationships = new_relationship.get_all()

    nr_delete_three = new_relationship.relate(nc_one, nc_three)

    assert len(new_relationship.get_all()) == len(new_relationships) + 1
    nr_delete_three.delete()
    assert len(new_relationship.get_all()) == len(new_relationships)

    df_rs = new_relationship.get_all().as_dataframe()

    p = DataPackage('test-csv', package_type='Tabular')
    dataset.add(p)
    dataset.update()
    assert p.exists

    p.relate_to(nc_one)
    p.relate_to(nc_two)
    nc_three.relate_to(p, new_relationship)
    new_relationship.relate(nc_four, p)

    assert len(nc_four.get_related(new_model.type)) == 5

def test_simple_model_properties(dataset):

     # Define properties as tuples
     model_with_basic_props_1 = dataset.create_model('Basic_Props_1', description='a new description',schema=
     [('an_integer', int, 'An Integer', True), ('a_bool', bool), ('a_string', str), ('a_date', datetime.datetime)])

     assert dataset.get_model(model_with_basic_props_1.id) == model_with_basic_props_1

     # Add a property with a description
     model_with_basic_props_1.add_property('a_new_property', float, display_name='Weight', description="some metric")

     updated_model_1 = dataset.get_model(model_with_basic_props_1.id)

     test_prop = updated_model_1.get_property('a_new_property')
     assert test_prop.display_name == 'Weight'
     assert test_prop.description == "some metric"

     # Define properties as ModelProperty objects
     model_with_basic_props_2 = dataset.create_model('Basic_Props_2', description='a new description', schema=[
         ModelProperty('name', data_type=str, title=True, required=True),
         ModelProperty('age', data_type=int),
         ModelProperty('DOB', data_type=datetime.datetime)
     ])

     assert dataset.get_model(model_with_basic_props_2.id) == model_with_basic_props_2

     # Add a property
     model_with_basic_props_2.add_property('weight2', ModelPropertyType(data_type=float), display_name='Weight')

     updated_model_2 = dataset.get_model(model_with_basic_props_2.id)

     assert updated_model_2.get_property('weight2').display_name == 'Weight'
     assert updated_model_2.get_property('name').required == True

     # Define properties as ModelProperty objects with ModelPropertyType data_type
     model_with_basic_props_3 = dataset.create_model('Basic_Props_3', description='a new description', schema=[
         ModelProperty('name', data_type=ModelPropertyType(data_type=str), title=True),
         ModelProperty('age', data_type=ModelPropertyType(data_type=int)),
         ModelProperty('DOB', data_type=ModelPropertyType(data_type=str))
     ])

     assert dataset.get_model(model_with_basic_props_3.id) == model_with_basic_props_3

     # Add a property
     model_with_basic_props_3.add_property('weight3', ModelPropertyType(data_type=float), display_name='Weight')

     updated_model_3 = dataset.get_model(model_with_basic_props_3.id)

     assert updated_model_3.get_property('weight3').display_name == 'Weight'

     # Reverse look up property data types
     ModelProperty('name', data_type='string', title=True, required=True)
     model_with_basic_props_4 = dataset.create_model('Basic_Props_4', description='a new description', schema=[
         ModelProperty('name', data_type='string', title=True, required=True),
         ModelProperty('age', data_type='long'),
         ModelProperty('DOB', data_type='date')
     ])

     assert dataset.get_model(model_with_basic_props_4.id) == model_with_basic_props_4

     # Add a property
     model_with_basic_props_4.add_property('weight4', ModelPropertyType(data_type='double'), display_name='Weight')

     updated_model_4 = dataset.get_model(model_with_basic_props_4.id)

     assert updated_model_4.get_property('weight4').display_name == 'Weight'
     assert updated_model_4.get_property('name').required == True

def test_complex_model_properties(dataset):
    model_with_complex_props = dataset.create_model('Complex_Props', description='a new description', schema=[
        ModelProperty('name', data_type=ModelPropertyType(data_type=str), title=True),
        ModelProperty('age', data_type=ModelPropertyType(data_type=int)),
        ModelProperty('email', data_type=ModelPropertyType(data_type=str, format='email'))
    ])

    assert dataset.get_model(model_with_complex_props.id) == model_with_complex_props

    # Add a property
    model_with_complex_props.add_property('weight', ModelPropertyType(data_type=float, unit='kg'), display_name='Weight')

    updated_model = dataset.get_model(model_with_complex_props.id)

    weight_property = updated_model.get_property('weight')
    assert (weight_property.display_name == 'Weight')
    assert (weight_property.type == float)
    assert (weight_property._type.data_type == float)
    assert (weight_property._type.format == None)
    assert (weight_property.unit == 'kg')

    email_property = model_with_complex_props.get_property('email')
    assert (email_property.type == unicode)
    assert (email_property._type.data_type == unicode)
    assert (email_property._type.format.lower() == 'email')
    assert (email_property.unit == None)

    good_values = {'name': 'Bob', 'age': 1, 'email': 'test@test.com', 'weight': 10}
    good_record = updated_model.create_record(good_values)

    bad_values = {'name': 'Bob', 'age': 1, 'email': '123455', 'weight': 10}

    with pytest.raises(Exception):
        bad_record = updated_model.create_record(bad_values)


def test_get_connected(dataset):
    model_1 = dataset.create_model('Model_A', description="model a", schema=[
        ModelProperty("prop1", data_type=ModelPropertyType(data_type=str),
                      title=True)])

    related_models = model_1.get_connected()
    # For a single, unconnected model, it should return nothing
    assert len(related_models) == 0

    model_2 = dataset.create_model('Model_B', description="model b", schema=[
        ModelProperty("prop1", data_type=ModelPropertyType(data_type=str),
                      title=True)])

    relationship = dataset.create_relationship_type(
        'New_Relationship_{}'.format(current_ts()), 'a new relationship')

    model_instance_1 = model_1.create_record({"prop1": "val1"})
    model_instance_2 = model_2.create_record({"prop1": "val1"})
    model_instance_1.relate_to(model_instance_2, relationship)

    related_models = model_1.get_connected()
    assert len(related_models) == 2

    # For a connected model, return all connections + the model itself
    related_models = model_2.get_connected()
    assert len(related_models) == 2

    # Check that get_connected_models from the dataset object also works
    related_models = dataset.get_connected_models(model_1.id)
    assert len(related_models) == 2

    related_models = dataset.get_connected_models(model_2.id)
    assert len(related_models) == 2


def test_array_model_properties(dataset):
    model = dataset.create_model('Array_Properties', description='Description', schema=[
        ModelProperty('name', data_type=str, title=True),
        ModelProperty('int_array', data_type=ModelPropertyEnumType(
            data_type=int, multi_select=True))
    ])

    int_array = model.get_property('int_array')
    assert int_array.type == int
    assert int_array.multi_select == True
    assert int_array.enum == None

    record = model.create_record({'name': 'A', 'int_array': [1, 2, 3]})
    bulk_records = model.create_records([
        {'name': 'B', 'int_array': [1, 2, 3]},
        {'name': 'B', 'int_array': [1, 2, 3]},
        {'name': 'B', 'int_array': [1, 2, 3]}
    ])

    gotten = model.get_all()
    assert all(x.values['int_array'] == [1, 2, 3] for x in gotten)


def test_enum_model_properties(dataset):
    model = dataset.create_model('Enum_Props', description='a new description', schema=[
        ModelProperty('name', data_type=str, title=True),
        ModelProperty('int_enum', data_type=ModelPropertyEnumType(
            data_type=float, enum=[1.0, 2.0, 3.0], unit="cm", multi_select=False)),
        ModelProperty('str_enum', data_type=ModelPropertyEnumType(
            data_type=str, enum=['foo', 'bar', 'baz'], multi_select=True)),
    ])

    assert model == dataset.get_model(model.id)

    int_enum = model.get_property('int_enum')
    str_enum = model.get_property('str_enum')

    assert (int_enum.type == float)
    assert (int_enum.multi_select == False)
    assert (int_enum.enum == [1.0, 2.0, 3.0])
    assert (int_enum.unit == 'cm')

    assert (str_enum.type == unicode)
    assert (str_enum.multi_select == True)
    assert (str_enum.enum == ['foo', 'bar', 'baz'])

    record = model.create_record({'name': 'A', 'int_enum': 1.0, 'str_enum': ['foo', 'bar']})

    gotten = model.get_all()[0]
    assert gotten.values['int_enum'] == 1.0
    assert set(gotten.values['str_enum']) == set(['foo', 'bar'])

    with pytest.raises(Exception, match=r"Value '5' is not a member*"):
        model.create_record({'name': 'A', 'int_enum': 5, 'str_enum': ['foo', 'bar']})

    with pytest.raises(Exception, match=r"Value 'blah' is not a member*"):
        model.create_record({'name': 'A', 'int_enum': 1.0, 'str_enum': ['blah']})


Graph = namedtuple('TestGraph', ['dataset', 'models', 'model_records',
                                 'relationships', 'relationship_records'])


@pytest.fixture(scope="module")
def simple_graph(client):
    """
        Creates a small test graph in an independent dataset to de-couple
        from other tests
    """
    test_dataset = create_test_dataset(client)
    model_1 = test_dataset.create_model(
        'Model_A', description="model a", schema=[
            ModelProperty("prop1", data_type=ModelPropertyType(data_type=str),
                          title=True)])

    model_2 = test_dataset.create_model(
        'Model_B', description="model b",
        schema=[ModelProperty("prop1",
                              data_type=ModelPropertyType(data_type=str),
                              title=True)
                ])

    relationship = test_dataset.create_relationship_type(
        'New_Relationship_{}'.format(current_ts()), 'a new relationship')

    model_instance_1 = model_1.create_record({"prop1": "val1"})
    model_instance_2 = model_2.create_record({"prop1": "val1"})
    model_instance_1.relate_to(model_instance_2, relationship)

    graph = Graph(test_dataset, models=[model_1, model_2],
                  model_records=[model_instance_1, model_instance_2],
                  relationships=[relationship],
                  relationship_records=None)
    yield graph

    ds_id = test_dataset.id
    client._api.datasets.delete(test_dataset)
    all_dataset_ids = [x.id for x in client.datasets()]
    assert ds_id not in all_dataset_ids
    assert not test_dataset.exists


def test_get_related_models(simple_graph):
    related_models = simple_graph.models[0].get_related()
    assert len(related_models) == 1
    assert related_models[0].type == simple_graph.models[1].type


def test_related_records_pagination(dataset):
    patient = dataset.create_model(
        'patient', description="patient", schema=[
            ModelProperty("name", data_type=ModelPropertyType(data_type=str),
                          title=True)])

    visit = dataset.create_model(
        'visit', description="visit", schema=[
            ModelProperty("field", data_type=ModelPropertyType(data_type=str),
                          title=True)])

    attends = dataset.create_relationship_type('attends', 'an attendance')

    fred = patient.create_record({"name": "Fred"})
    visits = visit.create_records([{"field": str(i)} for i in range(200)])

    fred.relate_to(visits, attends)

    # Get all records
    gotten = fred.get_related()
    assert len(gotten) == 200
    assert [int(r.get("field")) for r in gotten] == list(range(200))

def test_stringified_boolean_values(dataset):
    ppatient = dataset.create_model(
    'potential_patient', description="potential patient", schema=[
        ModelProperty("name", data_type=ModelPropertyType(data_type=str),
                      title=True),
        ModelProperty("sick", data_type=ModelPropertyType(data_type=bool))])

    ppatient.create_record({"name": "Fred", "sick": "false"})
    ppatient.create_record({"name": "Joe", "sick": "true"})
    ppatient.create_record({"name": "Adele", "sick": "plop"})

    # Get all records
    gotten = ppatient.get_all()
    assert len(gotten) == 3
    assert gotten[0]._values["sick"].value == False
    assert gotten[1]._values["sick"].value == True
    assert gotten[2]._values["sick"].value == True


def test_get_topology(simple_graph):
    topology = simple_graph.dataset.get_topology()
    assert 'models' in topology
    assert len(topology['models']) == 2
    assert 'relationships' in topology
    assert len(topology['relationships']) == 1

def test_get_graph_summary(simple_graph):
    summary = simple_graph.dataset.get_graph_summary()
    expected_fields = ["modelCount", "modelRecordCount", "modelSummary",
                       "relationshipCount", "relationshipRecordCount",
                       "relationshipSummary", "relationshipTypeCount",
                       "relationshipTypeSummary"]
    for f in expected_fields:
        assert f in summary.keys()

    assert summary['modelCount'] == 2
    assert summary['modelRecordCount'] == 2
    assert summary['relationshipCount'] == 1
    assert summary['relationshipRecordCount'] == 1
    assert summary['relationshipTypeCount'] == 1

def test_simple_query(simple_graph):
    model_a = simple_graph.dataset.get_model('Model_A')
    model_b = simple_graph.dataset.get_model('Model_B')

    # Filter via a model name:
    assert len(model_a.query().filter("prop1", "eq", "val1").run()) == 1
    assert len(model_a.query().filter("prop1", "eq", "not-present").run()) == 0

    # Join via a model name:
    assert len(model_a.query().join("Model_B", ("prop1", "eq", "val1")).run()) == 1
    assert len(model_a.query().join("Model_B", ("prop1", "eq", "val2")).run()) == 0

    # Filter via a model instance:
    assert len(model_a.query().filter("prop1", "eq", "val1").run()) == 1
    assert len(model_a.query().filter("prop1", "eq", "not-present").run()) == 0

    # Join via a model instance:
    assert len(model_a.query().join(model_b, ("prop1", "eq", "val1")).run()) == 1
    assert len(model_a.query().join(model_b, ("prop1", "eq", "val2")).run()) == 0

    # use select, join, and filter:
    result = model_a.query() \
            .select(model_b) \
            .filter("prop1", "eq", "val1") \
            .join(model_b, ("prop1", "eq", "val1")) \
            .run()

    assert len(result) == 1
    assert len(result[0].items()) == 1
    assert result[0].target.type == model_a.type
    assert result[0].get(model_b).type == model_b.type

    # and with an offset of 1:
    result = model_a.query() \
            .select(model_b) \
            .filter("prop1", "eq", "val1") \
            .join(model_b, ("prop1", "eq", "val1")) \
            .offset(1) \
            .run()
    assert len(result) == 0

    # and with a limit of 0:
    result = model_a.query() \
            .select(model_b) \
            .filter("prop1", "eq", "val1") \
            .join(model_b, ("prop1", "eq", "val1")) \
            .limit(0) \
            .run()
    assert len(result) == 0



from __future__ import absolute_import, division, print_function
from future.utils import integer_types, string_types

import blackfynn.log as log
from blackfynn.models import get_package_class

# urllib compatibility
from future import standard_library
standard_library.install_aliases()
import urllib.parse


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Base class
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class APIBase(object):
    host = None
    base_uri = ''
    name = ''

    def __init__(self, session):
        """
        Base class to be used by all API components.
        """
        # api session
        self.session = session
        self._logger = log.get_logger('blackfynn.api')

    def _get_id(self, thing):
        """
        Get ID for object. Assumes string is already ID.
        """
        if isinstance(thing, (string_types, integer_types)):
            return thing
        elif thing is None:
            return None
        else:
            return thing.id

    def _get_int_id(self, thing):
        """
        Get internal ID for object.
        """
        if isinstance(thing, (string_types, integer_types)):
            return thing
        elif thing is None:
            return None
        else:
            return thing.int_id

    def _get_package_from_data(self, data):
        # parse json
        cls = get_package_class(data)
        pkg = cls.from_dict(data, api=self.session)

        return pkg

    def _uri(self, url_str, **kwvars):
        vals = {k:urllib.parse.quote(str(var)) for k,var in kwvars.items()}
        return url_str.format(**vals)

    def _get(self, endpoint, base=None, host=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        host = self.host     if host is None else host
        return self.session._call('get', endpoint, host=host, base=base, *args, **kwargs)

    def _post(self, endpoint, base=None, host=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        host = self.host     if host is None else host
        return self.session._call('post', endpoint, host=host, base=base, *args, **kwargs)

    def _put(self, endpoint, base=None, host=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        host = self.host     if host is None else host
        return self.session._call('put', endpoint, host=host, base=base, *args, **kwargs)

    def _del(self, endpoint, base=None, host=None, *args, **kwargs):
        base = self.base_uri if base is None else base
        host = self.host     if host is None else host
        return self.session._call('delete', endpoint, host=host, base=base, *args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from builtins import dict, object

import base64
import json

import requests
from requests import Session
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from requests.packages.urllib3.util.retry import Retry
from future.utils import raise_from

# blackfynn
import blackfynn.log as log
from blackfynn.models import User


class UnauthorizedException(Exception):
    pass


class BlackfynnRequest(object):
    def __init__(self, func, uri, *args, **kwargs):
        self._func = func
        self._uri = uri
        self._args = args
        self._kwargs = kwargs
        self._response = None

        self._logger = log.get_logger('blackfynn.base.BlackfynnRequest')

    def raise_for_status(self,resp):
        try:
            resp.raise_for_status()
        except HTTPError as e: #raise for status raise an HTTPError, so we can use it to grab the message
            if resp.text:
                raise_from(HTTPError(resp.content, response=resp), e)
            else:
                raise e
        return

    def _handle_response(self, resp):
        self._logger.debug(u"resp = {}".format(resp))
        self._logger.debug(u"resp.content = {}".format(resp.text)) # decoded unicode
        if resp.status_code in [requests.codes.forbidden, requests.codes.unauthorized]:
            raise UnauthorizedException()

        if not resp.status_code in [requests.codes.ok, requests.codes.created]:
            self.raise_for_status(resp)
        try:
            # return object from json
            resp.data = json.loads(resp.text)
        except:
            # if not json, still return response content
            resp.data = resp.text

    def call(self, timeout=None):
        self._response = self._func(self._uri, *self._args, timeout=timeout, **self._kwargs)
        self._handle_response(self._response)
        return self._response


class ClientSession(object):
    def __init__(self, settings):
        self._host = settings.api_host
        self._api_token = settings.api_token
        self._api_secret = settings.api_secret
        self._jwt = settings.jwt
        self._model_service_host = settings.model_service_host
        self._logger = log.get_logger('blackfynn.base.ClientSession')

        self._session = None
        self._token = None
        self._secret = None
        self._context = None
        self._organization = None
        self.profile = None
        self.settings = settings

    def authenticate(self, organization = None):
        """
        """
        if self._jwt is None:
            self._authenticate_with_session(organization = organization)
        else:
            self._authenticate_with_jwt(organization = organization)

    def _authenticate_with_session(self, organization = None):
        """
        An API token is used to authenticate against the Blackfynn platform.
        The token that is returned from the API call will be used for all
        subsequent API calls.
        """
        # make authentication request
        session_response = self._post('/account/api/session', json=dict(tokenId = self._api_token, secret = self._api_secret))

        # parse response, set session
        self.token = session_response['session_token']
        self.profile = User.from_dict(self._get('/user/'))

        if organization is None:
            organization = session_response.get('organization')

        self._set_org_context(organization)

    def _authenticate_with_jwt(self, organization = None):
        """
        Use a JWT to make all subsequent requests to API.
        """
        # Force session creation:
        self.session
        self.token = self._jwt
        # Get the associated user and set the org context to the preferred
        # org set for the user. Additionally, if authentication fails at this
        # step don't attempt to reauthenticate recursively.
        user = self._get("/user", reauthenticate=False)
        organization = user['preferredOrganization']
        self._set_org_context(organization)

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, value):
        self._token = value
        self._set_auth(value)

    def _set_org_context(self, organization_id):
        self._organization = organization_id
        self._session.headers['X-ORGANIZATION-ID'] = organization_id

    def _set_auth(self, session_token):
        self._session.headers['Authorization'] = 'Bearer {}'.format(session_token)
        # If the JWT is present, we need to skip the `X-SESSION-ID` header
        # as it will cause the JWT to be interpreted as a regular session token.
        if self._jwt is None:
            self._session.headers['X-SESSION-ID'] = session_token

    @property
    def session(self):
        """
        Make requests-futures work within threaded/distributed environment.
        """
        if self._session is None:
            self._session = Session()
            self._set_auth(self.token)

            # Enable retries via urllib
            adapter = HTTPAdapter(
                max_retries=Retry(
                    total=self.settings.max_request_timeout_retries,
                    backoff_factor=.5,
                    status_forcelist=[502, 503, 504] # Retriable errors (but not POSTs)
                )
            )
            self._session.mount('http://', adapter)
            self._session.mount('https://', adapter)

        return self._session

    def _make_request(self, func, uri, *args, **kwargs):
        self._logger.debug('~'*60)
        self._logger.debug("uri = {} {}".format(func.__func__.__name__, uri))
        self._logger.debug("args = {}".format(args))
        self._logger.debug("kwargs = {}".format(kwargs))
        self._logger.debug("headers = {}".format(self.session.headers))
        return BlackfynnRequest(func, uri, *args, **kwargs)

    def _call(self, method, endpoint, base='', reauthenticate=True, *args, **kwargs):
        if method == 'get':
            func = self.session.get
        elif method == 'put':
            func = self.session.put
        elif method == 'post':
            func = self.session.post
        elif method == 'delete':
            func = self.session.delete

        # serialize data
        if 'data' in kwargs:
            kwargs['data'] = json.dumps(kwargs['data'])

        # we might specify a different host
        if 'host' in kwargs:
            host = kwargs['host']
            kwargs.pop('host')
        else:
            host = self._host

        # call endpoint
        uri = self._uri(endpoint, base=base, host=host)
        req = self._make_request(func, uri, *args, **kwargs)
        resp = self._get_response(req, reauthenticate=reauthenticate)

        return resp.data

    def _uri(self, endpoint, base, host=None):
        if host is None:
            host = self._host
        return '{}{}{}'.format(host, base, endpoint)

    def _get(self, endpoint, *args, **kwargs):
        return self._call('get', endpoint, *args, **kwargs)

    def _post(self, endpoint, *args, **kwargs):
        return self._call('post', endpoint, *args, **kwargs)

    def _put(self, endpoint, *args, **kwargs):
        return self._call('put', endpoint, *args, **kwargs)

    def _del(self, endpoint, *args, **kwargs):
        return self._call('delete', endpoint, *args, **kwargs)

    def _get_response(self, req, reauthenticate=True):
        try:
            return req.call(timeout=self.settings.max_request_time)
        except UnauthorizedException as e:
            if self._token is None or reauthenticate is False:
                raise e

            # try to refresh the session and re-request
            self.authenticate(self._organization)
            return req.call(timeout=self.settings.max_request_time)

    def register(self, *components):
        """
        Register API component with session. Components should all be of
        APIBase type and have a name and base_uri property.

        The registered component will have reference to base session to
        make higher-level calls outside of its own scope, if needed.
        """
        # initialize
        for component in components:
            c = component(session=self)
            assert len(component.name) > 1, "Invalid API component name"
            # component is accessible via session.(name)
            self.__dict__.update({ component.name: c })

    @property
    def headers(self):
        return self.session.headers



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from blackfynn.api.base import APIBase


class UserAPI(APIBase):
    base_uri = "/user"
    name = 'user'

    def switch_organization(self, orgId):
        return self._put( self._uri('/organization/{orgId}/switch'.format(orgId=orgId)))



import logging
import os


def test_default_log_level():
    import blackfynn.log as log

    # default log level should be INFO
    base_log = log.get_logger('base_log')
    assert base_log.getEffectiveLevel() == logging.INFO


def test_log_level_based_on_env():
    import blackfynn.log as log

    # setting env var should change logging level
    os.environ['BLACKFYNN_LOG_LEVEL'] = 'WARN'
    warn_log = log.get_logger('warn_log')
    assert warn_log.getEffectiveLevel() == logging.WARN
